#ifndef LOW_LEVEL_NET_H
#define LOW_LEVEL_NET_H

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <errno.h>

/**
 * This class provides low level network communication via sockets
 */

///structure for buffer that contains data and current size of the data
struct buffer_handle {
  int current_size;
  void *pData;
};


#ifdef __cplusplus
extern "C"  {
#endif
  
  /**
   * Initializes buffer with the default size
   * \return buffer_handle buffer
   */
  struct buffer_handle initialize_buffer();
  /**
   * Frees the buffer
   * \param[in] buffer_handle buffer
   */
  void destroy_handle(struct buffer_handle);
  /**
   * Updates buffer size with the max of given size and current size
   * \param[in] bh buffer
   * \param[in] new_size new size of the buffer
   */
  void update_buffer_size(struct buffer_handle* bh, int new_size);
  /**
   * Reads the msg from socket
   * \param[in] bh buffer
   * \param[in] fd socket id
   */
  void* read_wire_message(struct buffer_handle *bh, int fd);
  /**
   * Writes the msg to the socket
   * \param[in] fd socket id
   * \param[in] buf buffer
   * \param[in] nLength length of the msg
   * \return success of write
   */
  int write_wire_message(int fd, const void* buf, int nLength);
  
#ifdef __cplusplus
}
#endif

#endif
