/**
 * \file server_conn.cxx
 */
#include <iostream>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include "server_conn.h"
#include "globals.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <string.h>

///Default constructor
Hazy_Server::Hazy_Server()  {
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE;
  
  if (getaddrinfo("localhost", "10003", &hints, &servinfo) != 0) {
    printf("Error at getaddrinfo() : %s\n", gai_strerror(rc));
    exit(1);
  }
  
  if ((sockfd = socket(servinfo->ai_family, servinfo->ai_socktype, servinfo->ai_protocol)) == -1) {
    printf("Error at socket() : %s\n", strerror(errno));
    freeaddrinfo(servinfo);
    exit(1);
  }
  
  if (bind(sockfd, servinfo->ai_addr, servinfo->ai_addrlen) == -1) {
    printf("Error at bind() : %s\n", strerror(errno));
    freeaddrinfo(servinfo);
    close(sockfd);
  }
  
  if (listen(sockfd, CON_NUM) == -1) {
    printf("Error at listen() : %s\n", strerror(errno));
    freeaddrinfo(servinfo);
    close(sockfd);
    exit(1);
  }
  
  
  /* Disable the Nagle (TCP No Delay) algorithm */
  
  int flag = 1;
  if(setsockopt( sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) ) == -1) {
    printf("Couldn't setsockopt(TCP_NODELAY)\n");
    exit( EXIT_FAILURE );
  }
  
  sin_size = sizeof their_addr;
  bh = initialize_buffer();
}

/**
 * Accepts connection
 */
Hazy_Worker*
Hazy_Server::acceptConnection() {
  int new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);
  if(new_fd != -1) {
    return new Hazy_Worker(new_fd);
  } else { 
    return NULL;
  }
}

/**
 * Initialize connection
 */
void
Hazy_Server::inetConn() {
  inet_ntop(their_addr.sin_family, (struct sockaddr*)&their_addr.sin_addr, ip, sizeof ip);
}


/**
 * Closes server connection
 */
void
Hazy_Server::closeConnection() {
  freeaddrinfo(servinfo);
  close(sockfd);
}

/**
 * Reads data from client
 */
void*
Hazy_Worker::readData() {
  return read_wire_message(&bh, fd);  
}

/**
 * Writes data to the client
 * \param[in] data data pointer to be written to the client
 * \param[in] size size of the data pointer
 * \param[in] intData boolean value shows whether data is integer data or not
 */
int
Hazy_Worker::writeData(const void *data, int sizeInBytes, bool intData) {
  VERBOSE_ONLY(std::cout << "[writeData]" << "writing " << sizeInBytes << "bytes the answer is " << (const char*) data << std::endl;);
  int  numOfBytesSent = write_wire_message(fd, (const char*) data, sizeInBytes);
  if(numOfBytesSent < 0) {
    printf("[ERROR:Send]  %s\n", strerror(errno));
    return numOfBytesSent;
  }
  return numOfBytesSent;
}

/**
 * Closes client connection
 */
void
Hazy_Worker::closeClientConnection() {
  close(fd);
}
