#include "create_view.h"

Hazy_Create_View::Hazy_Create_View(std::string _db_name, std::string _view_name, 
				   std::string _view_key, std::string _entity_table, 
				   std::string _entity_table_key, std::string _labels_table, 
				   std::string _label_in_table, std::string _example_table, 
				   std::string _example_table_key, std::string _example_table_label, 
				   std::string _feature_function, enum storage_manager_type _sm, 
				   int _dense, int _useepsmap, hazy_model::strategy _hazy_strategy, 
				   int _reservoir_size, double _resort_time, double _tao_val, 
				   double _alpha_ratio, int _buffer_size) {
  this->hazy_db_name = _db_name;
  this->view_name = _view_name;
  this->view_key = _view_key;
  this->original_entity_table = _entity_table;
  this->entity_table_key = _entity_table_key;
  this->labels_table = _labels_table;
  this->label_in_table = _label_in_table;
  this->example_table = _example_table;
  this->example_table_key = _example_table_key;
  this->example_table_label = _example_table_label;
  this->feature_function = _feature_function;
  this->sm = _sm;
  this->dense = _dense;
  this->useepsmap = _useepsmap;
  this->hazy_strategy = _hazy_strategy;
  this->reservoir_size = _reservoir_size;
  this->resort_time = _resort_time;
  this->tao_val = _tao_val;
  this->alpha_ratio = _alpha_ratio;
  this->buffer_size = _buffer_size;
  
  db_conn = new Hazy_Database(hazy_db_name);
}


const std::string error_msg = "There is an error in the query\n";

void
Hazy_Create_View::retrieve_unique_view_id(int &view_id) {
  int count = 0;
  std::ostringstream query;
  query << "SELECT COUNT(*) FROM " << CLASSIFICATION_GENERAL_TABLE << ";";
  int ret_val = db_conn->execute_query_msg_int(query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, query.str().c_str());
  
  if(count == 0)
    view_id = 0;
  else {
    std::ostringstream query;
    query << "SELECT MAX(" << VIEW_ID << ") FROM " << CLASSIFICATION_GENERAL_TABLE << ";";
    int ret_val = db_conn->execute_query_msg_int(query.str().c_str(), error_msg.c_str(), view_id);
    checkQueryReturnValue(ret_val, query.str().c_str());
    //get unique view id
    view_id ++;
  }

  std::ostringstream training_table_name;
  training_table_name << TRAINING_TABLE << view_id;
  this->training_table_name = training_table_name.str();
  std::ostringstream hazy_entity_table_name;
  hazy_entity_table_name << HAZY_ENTITY_TABLE_BASE_NAME << view_id;
  this->hazy_entity_table = hazy_entity_table_name.str();
}

void
Hazy_Create_View::insert_newview_to_catalog(int view_id) {
  std::ostringstream insert_query;
  insert_query << "INSERT INTO " << CLASSIFICATION_GENERAL_TABLE
	       << "(" << VIEW_ID << ", " << VIEW_NAME << ", " << ENTITY_TABLE_NAME << ", " << LABELS_TABLE_NAME
	       << ", " << EXAMPLE_TABLE_NAME << ", " << FEATURE_FUNCTION << ") "
	       << "VALUES (" << view_id << ", '" << view_name << "', '" << original_entity_table << "', '" << labels_table
	       << "', '" << example_table << "', '" << feature_function << "')";
  
  int ret_val = db_conn->execute_statement_msg(insert_query.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, insert_query.str().c_str());
}

void
Hazy_Create_View::callGetStatsFeaturesFunction(int view_id) {
  std::ostringstream query;
  query << "SELECT * FROM " << GET_STATS_FEATURES_FUNCTION_NAME << "(" << view_id
	<< ", '" << original_entity_table << "');";
  int ret_val = db_conn->execute_statement_msg(query.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, query.str().c_str());
}

void
Hazy_Create_View::normalizeAndGetM(int view_id, double &m_factor) {
  std::ostringstream query;
  query << "SELECT * FROM " << NORMALIZE_GET_M_FACTOR_FUNCTION_NAME
	<< "(" << view_id << ");";
  int ret_val = db_conn->execute_query_msg_double(query.str().c_str(), error_msg.c_str(), m_factor);
  checkQueryReturnValue(ret_val, query.str().c_str());
  this->m_factor = m_factor;
}

void
Hazy_Create_View::callCreateEntityFeaturesFunction(int view_id) {
  std::ostringstream query;
  query << "SELECT * FROM " << CREATE_ENTITY_FEATURES_FUNCTION_NAME << " (" << view_id << ");";
  int ret_val = db_conn->execute_statement_msg(query.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, query.str().c_str());
}

void
Hazy_Create_View::callPrepareTrainFunction(int view_id) {
  std::ostringstream query;
  query << "SELECT * FROM " << PREPARE_TRAIN_FUNCTION_NAME << " (" << view_id << ", '"
	<< example_table << "', '" << example_table_key << "', '" << example_table_label << "');";
  int ret_val = db_conn->execute_statement_msg(query.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, query.str().c_str());
}

void
Hazy_Create_View::getNumberOfFeatures(int view_id, int &num_of_features) {
  std::ostringstream query;
  query << "SELECT COUNT(*) FROM " << FEATURE_IDS_TABLE_BASE_NAME << view_id << ";";
  int ret_val = db_conn->execute_query_msg_int(query.str().c_str(), error_msg.c_str(), num_of_features);
  checkQueryReturnValue(ret_val, query.str().c_str());

  //NB: dim = getNumberOfFeatures + 1
  this->dim = num_of_features + 1;
}

/**
 * Writes the given msg to the server and retrieves integer from the received msg
 * \param[in] msg msg to be written to the server
 * \param[in] length size of the msg
 */
int
Hazy_Create_View::write_func_create_view(message_data msg, size_t length) {
  struct addrinfo *res = NULL;
  int sockfd = -1;
  pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;
  pthread_mutex_lock( &mutex1 );
  
  int buffer_initialized = 0;
  struct buffer_handle update_trig_buf;
  if (buffer_initialized == 0) { update_trig_buf = initialize_buffer(); buffer_initialized = 1; } 
  
  struct addrinfo hints;
  
  if(sockfd < 0) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      std::cerr << "[create_view.cxx] Getaddrinfo fails\n" << std::endl;
      res = NULL;
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      return -1;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    
    if(sockfd < 0) {
      std::cerr << "[create_view.cxx] msg error: %s \n" << std::endl;
      sockfd = -1;
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      
      std::cerr << "[create_view.cxx] Connect failed!" << std::endl;
      return -1;
    }
    
    int flag = 1;
    if(setsockopt( sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) ) == -1) {
      std::cerr << "Couldn't setsockopt(TCP_NODELAY)\n" << std::endl;
      destroy_handle(update_trig_buf);
      pthread_mutex_unlock( &mutex1 );
      std::cerr << "EXIT_FAILURE" << std::endl;
    }
    
    if(connect(sockfd, res->ai_addr, res->ai_addrlen) < 0) {
      std::cerr << "[create_view.cxx] msg error: %s \n" << std::endl;
      sockfd = -1;
      pthread_mutex_unlock( &mutex1 );
      
      std::cerr << "[create_view.cxx] Connect failed\n" << std::endl;
      return -1;
    }
  }

  if(write_wire_message(sockfd, msg, length) != (int) length) {
    std::cerr << "msg length: " << length << std::endl;
    std::cerr << "msg error: " << strerror(errno) << std::endl;
    sockfd = -1;
    destroy_handle(update_trig_buf);
    pthread_mutex_unlock( &mutex1 );
    
    std::cerr << "[create_view.cxx] Write failed!" << std::endl;
    return -1;
  }	
  
  char *pData = (char *) read_wire_message(&update_trig_buf, sockfd);
  
  if(pData == NULL) {
    std::cerr << "msg error: " << strerror(errno) << std::endl;
    sockfd = -1;
    destroy_handle(update_trig_buf);
    pthread_mutex_unlock( &mutex1 );
    
    std::cerr << "[create_view.cxx] Read Failed!" << std::endl;
    return -1;
  }
  
  pthread_mutex_unlock( &mutex1 );
  
  return 0;
}

std::string
Hazy_Create_View::create_one_param_func_str(int view_id, std::string fnc_name,
					    std::string param, std::string return_type,
					    std::string so_file_source_base_name,
					    std::string source_fnc_name) {
  std::ostringstream fnc_str;
  fnc_str << "DROP FUNCTION IF EXISTS " << fnc_name << view_id << "(" << param 
	  << "); CREATE OR REPLACE FUNCTION " << fnc_name << view_id << "(" 
	  << param << ") RETURNS " << return_type << " AS '" 
	  << so_file_source_base_name << view_id << "." << SHARED_OBJECT_FILE_EXT 
	  << "', '" << source_fnc_name << view_id << "' LANGUAGE C;";

  return fnc_str.str();
}

int
Hazy_Create_View::createBackendFunctions(int view_id) {
  std::string src_dir_path = retrieve_hazy_dirs_path();
  std::string hazy_conf_file_path = retrieve_hazy_conf_file_path();
  std::string function_dir_path = src_dir_path + "/" + INSTALLATION_DIR + 
    "/" + FUNCTIONS_DIR;
  std::string original_backend_fnc_file_name = function_dir_path + "/" + BACKEND_FNC_FILE_NAME +
    + "." + SOURCE_CODE_EXT;
  
  size_t found = original_backend_fnc_file_name.find_last_of(".");
  std::ostringstream new_backend_fnc_file_name;
  new_backend_fnc_file_name << original_backend_fnc_file_name.substr(0, found) << view_id << "." 
				   << SOURCE_CODE_EXT;

  std::string line = "";

  std::ifstream original_file(original_backend_fnc_file_name.c_str());
  std::ofstream new_file(new_backend_fnc_file_name.str().c_str());


  if(original_file.is_open()) {
    while(original_file.good()) {
      getline(original_file, line);
      bool line_changed;

      std::ostringstream model_array_str;
      model_array_str << "model_array" << view_id;
      std::ostringstream waste_count_str;
      waste_count_str << "waste_count" << view_id;
      std::ostringstream not_waste_count_str;
      not_waste_count_str << "not_waste_count" << view_id;
      std::ostringstream bias_str;
      bias_str << "bias" << view_id;
      std::ostringstream dot_prdct_sparse_cached_waste;
      dot_prdct_sparse_cached_waste << DOT_PRDCT_SPARSE_CACHED_WASTE << view_id;
      std::ostringstream dot_prdct_sparse_cached_without_waste;
      dot_prdct_sparse_cached_without_waste << DOT_PRDCT_SPARSE_CACHED_WITHOUT_WASTE << view_id;
      std::ostringstream dot_prdct_dense_cached_waste;
      dot_prdct_dense_cached_waste << DOT_PRDCT_DENSE_CACHED_WASTE << view_id;
      std::ostringstream dot_prdct_dense_cached_without_waste;
      dot_prdct_dense_cached_without_waste << DOT_PRDCT_DENSE_CACHED_WITHOUT_WASTE << view_id;
      std::ostringstream initialize_waste_count_fnc;
      initialize_waste_count_fnc << INITIALIZE_WASTE_COUNT << view_id;
      std::ostringstream get_waste_ratio_fnc;
      get_waste_ratio_fnc << GET_WASTE_RATIO << view_id;
      std::ostringstream cache_model_fnc;
      cache_model_fnc << CACHE_MODEL << view_id;

      do {
	line_changed = false;

	std::string search_str = "";
	if(line.find("model_array") != std::string::npos &&
	   line.find("model_array") != line.find(model_array_str.str().c_str()))
	  search_str = "model_array";
	else if(line.find("waste_count") != std::string::npos &&
		line.find("waste_count") != line.find(waste_count_str.str().c_str()))
	  search_str = "waste_count";
	else if(line.find("not_waste_count") != std::string::npos &&
		line.find("not_waste_count") != line.find(not_waste_count_str.str().c_str()))
	  search_str = "not_waste_count";
	else if(line.find("bias") != std::string::npos &&
		line.find("bias") != line.find(bias_str.str().c_str()))
	  search_str = "bias";
	else if(line.find(DOT_PRDCT_SPARSE_CACHED_WASTE) != std::string::npos &&
		line.find(DOT_PRDCT_SPARSE_CACHED_WASTE) != line.find(dot_prdct_sparse_cached_waste.str().c_str()))
	  search_str = DOT_PRDCT_SPARSE_CACHED_WASTE;
	else if(line.find(DOT_PRDCT_SPARSE_CACHED_WITHOUT_WASTE) != std::string::npos &&
		line.find(DOT_PRDCT_SPARSE_CACHED_WITHOUT_WASTE) != line.find(dot_prdct_sparse_cached_without_waste.str().c_str()))
	  search_str = DOT_PRDCT_SPARSE_CACHED_WITHOUT_WASTE;
	else if(line.find(DOT_PRDCT_DENSE_CACHED_WASTE) != std::string::npos &&
		line.find(DOT_PRDCT_DENSE_CACHED_WASTE) != line.find(dot_prdct_dense_cached_waste.str().c_str()))
	  search_str = DOT_PRDCT_DENSE_CACHED_WASTE;
	else if(line.find(DOT_PRDCT_DENSE_CACHED_WITHOUT_WASTE) != std::string::npos &&
		line.find(DOT_PRDCT_DENSE_CACHED_WITHOUT_WASTE) != line.find(dot_prdct_dense_cached_without_waste.str().c_str()))
	  search_str = DOT_PRDCT_DENSE_CACHED_WITHOUT_WASTE;
	else if(line.find(INITIALIZE_WASTE_COUNT) != std::string::npos &&
		line.find(INITIALIZE_WASTE_COUNT) != line.find(initialize_waste_count_fnc.str().c_str()))
	  search_str = INITIALIZE_WASTE_COUNT;
	else if(line.find(GET_WASTE_RATIO) != std::string::npos &&
		line.find(GET_WASTE_RATIO) != line.find(get_waste_ratio_fnc.str().c_str()))
	  search_str = GET_WASTE_RATIO;
	else if(line.find(CACHE_MODEL) != std::string::npos &&
		line.find(CACHE_MODEL) != line.find(cache_model_fnc.str().c_str()))
	  search_str = CACHE_MODEL;
    
	
	std::ostringstream search_str_str;
	search_str_str << search_str << view_id;
	if(line.find(search_str.c_str()) == line.find(search_str_str.str().c_str()))
	  search_str = "";
	
	if(search_str.length() != 0) {
	  size_t search_index = line.find(search_str.c_str());
	  std::ostringstream new_str;
	  new_str << line.substr(0, search_index) << search_str << view_id << line.substr(search_index + search_str.length());
	  line = new_str.str();
	  line_changed = true;
	}
      } while(line_changed);
    
      new_file << line << std::endl;

    }
    original_file.close();
    new_file.close();
  }
  else {
    std::cerr << "unable to open " << BACKEND_FNC_FILE_NAME << "." << SOURCE_CODE_EXT << " file" << std::endl;
    return -1;
  }

  //compile and install new original_update_trig_file
  //read postgres_path
  std::string postgres_path = "";

  std::ifstream hazy_conf_file(hazy_conf_file_path.c_str());
  if(hazy_conf_file.is_open()) {
    while(hazy_conf_file.good()) {
      getline(hazy_conf_file, line);
      if(line.find(POSTGRES_PATH_LINE_IN_CONF_FILE) != std::string::npos) {
	size_t eq_found = line.find_first_of("=");
	postgres_path = line.substr(eq_found + 1);
      }
    }
  }

  std::string s_libs_functions_path = src_dir_path + "/" + INSTALLATION_DIR + "/" + S_LIBS_DIR + "/" + FUNCTIONS_DIR;

  std::ostringstream compile_new_function;
  compile_new_function << GCC << " -I " << postgres_path << "/" << POSTGRES_INCLUDE 
		       << "/" << POSTGRES_SERVER << "  -fPIC -o " << s_libs_functions_path 
		       << "/" << BACKEND_FNC_NAME << view_id << "." << OBJECT_FILE_EXT << " -c -I " 
		       << src_dir_path << "/" << INCLUDE_DIR << "/ -I " << src_dir_path 
		       << "/" << COMM_SERVER_DIR << "/ -I " << src_dir_path << "/" 
		       << SVM_CODE_DIR << "/ " << new_backend_fnc_file_name.str();

  std::ostringstream create_new_shared_file;
  create_new_shared_file << GCC << " -shared -o " << s_libs_functions_path << "/" 
			 << BACKEND_FNC_NAME << view_id << "." << SHARED_OBJECT_FILE_EXT 
			 << " -I " << src_dir_path << "/" << INCLUDE_DIR 
			 << "/ -I " << src_dir_path << "/" << COMM_SERVER_DIR 
			 << "/ -I " << src_dir_path << "/" << SVM_CODE_DIR 
			 << "/ " << s_libs_functions_path << "/" << BACKEND_FNC_NAME 
			 << view_id << "." << OBJECT_FILE_EXT;

  //compile new trigger
  system(compile_new_function.str().c_str());
  system(create_new_shared_file.str().c_str());

  //do copying
  std::ostringstream copy_command;
  copy_command << "cp " << s_libs_functions_path << "/" << BACKEND_FNC_NAME << view_id << "." << SHARED_OBJECT_FILE_EXT 
	       << " " << postgres_path << "/" << POSTGRES_LIB;

  system(copy_command.str().c_str());

  std::string dot_cached_waste_pair = create_one_param_func_str(view_id, DOT_PRDCT_CACHED_WASTE, 
								PAIR_ARRAY_TYPE, FLOAT8_TYPE, BACKEND_FNC_NAME, 
								DOT_PRDCT_SPARSE_CACHED_WASTE);
  std::string dot_cached_without_waste_pair = create_one_param_func_str(view_id, DOT_PRDCT_CACHED_WIHOUT_WASTE, 
									PAIR_ARRAY_TYPE, FLOAT8_TYPE, BACKEND_FNC_NAME, 
									DOT_PRDCT_SPARSE_CACHED_WITHOUT_WASTE);
  std::string dot_cached_waste_float = create_one_param_func_str(view_id, DOT_PRDCT_CACHED_WASTE, 
								 FLOAT8_ARRAY_TYPE, FLOAT8_TYPE, BACKEND_FNC_NAME, 
								 DOT_PRDCT_DENSE_CACHED_WASTE);
  std::string dot_cached_without_waste_float = create_one_param_func_str(view_id, DOT_PRDCT_CACHED_WIHOUT_WASTE, 
									 FLOAT8_ARRAY_TYPE, FLOAT8_TYPE, 
									 BACKEND_FNC_NAME, DOT_PRDCT_DENSE_CACHED_WITHOUT_WASTE);

  std::ostringstream initialize_waste_count;
  initialize_waste_count << "DROP FUNCTION IF EXISTS " << INITIALIZE_WASTE_COUNT << view_id 
			 << "(); CREATE OR REPLACE FUNCTION " << INITIALIZE_WASTE_COUNT << view_id 
			 << "() RETURNS VOID AS '" << BACKEND_FNC_NAME << view_id << "." 
			 << SHARED_OBJECT_FILE_EXT << "', '" << INITIALIZE_WASTE_COUNT << view_id << "' LANGUAGE C;";

  std::ostringstream get_waste_ratio;
  get_waste_ratio << "DROP FUNCTION IF EXISTS " << GET_WASTE_RATIO << view_id << "(); CREATE OR REPLACE FUNCTION " 
		  << GET_WASTE_RATIO << view_id << "() RETURNS " << FLOAT8_TYPE << " AS '" 
		  << BACKEND_FNC_NAME << view_id << "." << SHARED_OBJECT_FILE_EXT
		  << "', '" << GET_WASTE_RATIO << view_id << "' LANGUAGE C;";

  std::ostringstream cache_model;
  cache_model << "DROP FUNCTION IF EXISTS " << CACHE_MODEL << view_id << "(" << FLOAT8_ARRAY_TYPE 
	      << ", " << FLOAT8_TYPE << "); CREATE OR REPLACE FUNCTION " << CACHE_MODEL << view_id 
	      << "(" << FLOAT8_ARRAY_TYPE << ", " << FLOAT8_TYPE << ") RETURNS VOID AS '" 
	      << BACKEND_FNC_NAME << view_id << "." << SHARED_OBJECT_FILE_EXT << "', '" 
	      << CACHE_MODEL << view_id << "' LANGUAGE C;";

  int ret_val = db_conn->execute_statement_msg(dot_cached_waste_pair.c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, dot_cached_waste_pair.c_str());
  ret_val = db_conn->execute_statement_msg(dot_cached_without_waste_pair.c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, dot_cached_without_waste_pair.c_str());
  ret_val = db_conn->execute_statement_msg(dot_cached_waste_float.c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, dot_cached_waste_float.c_str());
  ret_val = db_conn->execute_statement_msg(dot_cached_without_waste_float.c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, dot_cached_without_waste_float.c_str());
  ret_val = db_conn->execute_statement_msg(initialize_waste_count.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, initialize_waste_count.str().c_str());
  ret_val = db_conn->execute_statement_msg(get_waste_ratio.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, get_waste_ratio.str().c_str());
  ret_val = db_conn->execute_statement_msg(cache_model.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, cache_model.str().c_str());


  return 0;
}

int
Hazy_Create_View::createUpdateInPlaceFunction(int view_id) {
  char buf[MAX_PATH_LENGTH];
  readlink("/proc/self/exe", buf, sizeof(buf) - 1);
  std::string executable_path = buf;
  size_t found;
  found = executable_path.find_last_of("/");
  std::string src_dir_path = executable_path.substr(0, found) + "/..";
  std::string hazy_conf_file_path = src_dir_path + "/../" + HAZY_CONF_FILE;
  std::string function_dir_path = src_dir_path + "/" + INSTALLATION_DIR + 
    "/" + FUNCTIONS_DIR;
  std::string original_inplace_update_fnc_file_name = function_dir_path + "/" + INPLACE_UPDATE_FNC_FILE_NAME +
    + "." + SOURCE_CODE_EXT;

  found = original_inplace_update_fnc_file_name.find_last_of(".");
  std::ostringstream new_inplace_update_fnc_file_name;
  new_inplace_update_fnc_file_name << original_inplace_update_fnc_file_name.substr(0, found) << view_id << "." 
			    << SOURCE_CODE_EXT;

  //read original_update_trig_file_name and replace view_id = 0 with view_id = view_id
  std::string line = "";

  
  std::ifstream original_file(original_inplace_update_fnc_file_name.c_str());
  std::ofstream new_file(new_inplace_update_fnc_file_name.str().c_str());


  if(original_file.is_open()) {
    while(original_file.good()) {
      getline(original_file, line);
      if(line.find("static char *relation_name = \"sm_test_external_table\";") != std::string::npos) {
	std::ostringstream line_replace;
	line_replace << "static char *relation_name" << view_id << " = \"" << this->view_name << "\";";
	line = line_replace.str();
      }

      bool line_changed;

      std::ostringstream szDebug_str;
      szDebug_str << "szDebug" << view_id;
      std::ostringstream update_hack_str;
      update_hack_str << "update_hack" << view_id;
      std::ostringstream relation_name_str;
      relation_name_str << "relation_name" << view_id;
      std::ostringstream bNull_str;
      bNull_str << "bNull" << view_id;
      std::ostringstream bIsReplace_str;
      bIsReplace_str << "bIsReplace" << view_id;
      std::ostringstream vals_str;
      vals_str << "vals" << view_id;
      std::ostringstream UPDATE_INDEX_str;
      UPDATE_INDEX_str << "UPDATE_INDEX" << view_id;

      do {
	line_changed = false;

	std::string search_str = "";
	if(line.find("szDebug") != std::string::npos &&
	   line.find("szDebug") != line.find(szDebug_str.str().c_str()))
	  search_str = "szDebug";
	else if(line.find("update_hack") != std::string::npos &&
	   line.find("update_hack") != line.find(update_hack_str.str().c_str()))
	  search_str = "update_hack";
	else if(line.find("relation_name") != std::string::npos &&
	   line.find("relation_name") != line.find(relation_name_str.str().c_str()))
	  search_str = "relation_name";
	else if(line.find("bNull") != std::string::npos &&
	   line.find("bNull") != line.find(bNull_str.str().c_str()))
	  search_str = "bNull";
	else if(line.find("bIsReplace") != std::string::npos &&
	   line.find("bIsReplace") != line.find(bIsReplace_str.str().c_str()))
	  search_str = "bIsReplace";
	else if(line.find("vals") != std::string::npos &&
	   line.find("vals") != line.find(vals_str.str().c_str()))
	  search_str = "vals";
	else if(line.find("UPDATE_INDEX") != std::string::npos &&
	   line.find("UPDATE_INDEX") != line.find(UPDATE_INDEX_str.str().c_str()))
	  search_str = "UPDATE_INDEX";
	
	std::ostringstream search_str_str;
	search_str_str << search_str << view_id;
	if(line.find(search_str.c_str()) == line.find(search_str_str.str().c_str()))
	  search_str = "";
	
	if(search_str.length() != 0) {
	  size_t search_index = line.find(search_str.c_str());
	  std::ostringstream new_str;
	  new_str << line.substr(0, search_index) << search_str << view_id << line.substr(search_index + search_str.length());
	  line = new_str.str();
	  line_changed = true;
	}
      } while(line_changed);
    
      new_file << line << std::endl;
    }
    original_file.close();
    new_file.close();
  }
  else {
    std::cerr << "unable to open update_trig.c file" << std::endl;
    return -1;
  }

  //compile and install new original_update_trig_file
  //read postgres_path
  std::string postgres_path = "";

  std::ifstream hazy_conf_file(hazy_conf_file_path.c_str());
  if(hazy_conf_file.is_open()) {
    while(hazy_conf_file.good()) {
      getline(hazy_conf_file, line);
      if(line.find(POSTGRES_PATH_LINE_IN_CONF_FILE) != std::string::npos) {
	size_t eq_found = line.find_first_of("=");
	postgres_path = line.substr(eq_found + 1);
      }
    }
  }

  std::string s_libs_functions_path = src_dir_path + "/" + INSTALLATION_DIR + "/" + S_LIBS_DIR + "/" + FUNCTIONS_DIR;

  std::ostringstream compile_new_trigger;
  compile_new_trigger << GCC << " -I " << postgres_path << "/" << POSTGRES_INCLUDE << "/" 
		      << POSTGRES_SERVER << " -fPIC -o " << s_libs_functions_path << "/" 
		      << INPLACE_UPDATE_FNC_FILE_NAME << view_id << "." << OBJECT_FILE_EXT 
		      << " -c " << new_inplace_update_fnc_file_name.str();

  std::ostringstream create_new_shared_file;
  create_new_shared_file << GCC << " -shared -o " 
			 << s_libs_functions_path << "/" << INPLACE_UPDATE_FNC_FILE_NAME << view_id 
			 << "." << SHARED_OBJECT_FILE_EXT << " " 
			 << s_libs_functions_path << "/" << INPLACE_UPDATE_FNC_FILE_NAME << view_id 
			 << "." << OBJECT_FILE_EXT;


  //compile new trigger
  system(compile_new_trigger.str().c_str());
  system(create_new_shared_file.str().c_str());

  //do copying
  std::ostringstream copy_command;
  copy_command << "cp " << s_libs_functions_path << "/" << INPLACE_UPDATE_FNC_FILE_NAME << view_id 
	       << "." << SHARED_OBJECT_FILE_EXT << " " << postgres_path << "/" << POSTGRES_LIB;

  system(copy_command.str().c_str());
  return 0;
}

void
Hazy_Create_View::createView(int view_id) {
  this->createBackendFunctions(view_id);
  this->createUpdateInPlaceFunction(view_id);

  buffer_handle bh;
  bh = initialize_buffer();
  struct create_view_bin_str *cv_struct = new struct create_view_bin_str();
  cv_struct->view_id = view_id;
  cv_struct->sm = this->sm;
  cv_struct->dense = this->dense;
  cv_struct->useepsmap = this->useepsmap;
  cv_struct->hazy_strategy = this->hazy_strategy;
  cv_struct->reservoir_size = this->reservoir_size;
  cv_struct->resort_time = this->resort_time;
  cv_struct->tao_val = this->tao_val;
  cv_struct->alpha_ratio = this->alpha_ratio;
  cv_struct->buffer_size = this->buffer_size;
  cv_struct->view_name = this->view_name;
  cv_struct->view_key = this->view_key;
  cv_struct->entity_table_name = this->hazy_entity_table;
  cv_struct->entity_table_key = this->entity_table_key;
  cv_struct->training_table_name = this->training_table_name;
  cv_struct->dim = this->dim;
  cv_struct->db_name = this->hazy_db_name;
  cv_struct->m_factor = this->m_factor;

  int msgSize        = sizeof(struct create_view_bin_str) + cv_struct->view_name.size() + cv_struct->view_key.size() + 
    cv_struct->entity_table_name.size() + cv_struct->entity_table_key.size() + cv_struct->training_table_name.size() +
    cv_struct->db_name.size() + 1000;

  update_buffer_size(&bh, msgSize);
  
  if(build_create_view_msg(cv_struct, bh.pData, msgSize) < 0) {
    destroy_handle(bh);
    std::cout << "Building create view message failed" << std::endl;
    return;
  }
  
  //send this message to the server
  int retVal = write_func_create_view(bh.pData, strlen((char *) bh.pData));
  if(retVal < 0) {
    std::cerr << "There is a problem with sending the data" << std::endl;
    return;
  }
  
  destroy_handle(bh);
  delete cv_struct;
}

int
Hazy_Create_View::createTriggers(int view_id) {
  char buf[MAX_PATH_LENGTH];
  readlink("/proc/self/exe", buf, sizeof(buf) - 1);
  std::string executable_path = buf;
  size_t found;
  found = executable_path.find_last_of("/");
  std::string src_dir_path = executable_path.substr(0, found) + "/..";
  std::string hazy_conf_file_path = src_dir_path + "/../" + HAZY_CONF_FILE;
  std::string trigger_dir_path = src_dir_path + "/" + INSTALLATION_DIR + 
    "/" + TRIGGERS_DIR;
  std::string original_update_trig_file_name = trigger_dir_path + "/" + UPDATE_TRIG_FILE_NAME +
    + "." + SOURCE_CODE_EXT;

  found = original_update_trig_file_name.find_last_of(".");
  std::ostringstream new_update_trig_file_name;
  new_update_trig_file_name << original_update_trig_file_name.substr(0, found) << view_id << "." 
			    << SOURCE_CODE_EXT;

  //read original_update_trig_file_name and replace view_id = 0 with view_id = view_id
  std::string line = "";

  
  std::ifstream original_file(original_update_trig_file_name.c_str());
  std::ofstream new_file(new_update_trig_file_name.str().c_str());


  if(original_file.is_open()) {
    while(original_file.good()) {
      getline(original_file, line);
      if(line.find("int view_id = 0;") != std::string::npos) {
	std::ostringstream line_replace;
	line_replace << "int view_id = " << view_id << ";";
	line = line_replace.str();
      }

      bool line_changed;

      std::ostringstream addrinfo_res_str;
      addrinfo_res_str << "addrinfo_res" << view_id;
      std::ostringstream sockfd_str;
      sockfd_str << "sockfd" << view_id;
      std::ostringstream mutex1_str;
      mutex1_str << "mutex1" << view_id;
      std::ostringstream buffer_initialized_str;
      buffer_initialized_str << "buffer_initialized" << view_id;
      std::ostringstream update_trig_buf_str;
      update_trig_buf_str << "update_trig_buf" << view_id;
      std::ostringstream update_mutex_str;
      update_mutex_str << "update_mutex" << view_id;
      std::ostringstream bh_str;
      bh_str << "bh" << view_id;
      std::ostringstream vec_buf_str;
      vec_buf_str << "vec_buf" << view_id;
      std::ostringstream update_trig_bin_str;
      update_trig_bin_str << "update_trig_bin" << view_id;
      std::ostringstream update_trig_str;
      update_trig_str << "update_trig" << view_id;

      do {
	line_changed = false;

	std::string search_str = "";
	
	if(line.find("addrinfo_res") != std::string::npos && 
	   line.find("addrinfo_res") != line.find(addrinfo_res_str.str().c_str()))
	  search_str = "addrinfo_res";
	else if(line.find("sockfd") != std::string::npos && 
		line.find("sockfd") != line.find(sockfd_str.str().c_str()))
	  search_str = "sockfd";
	else if(line.find("mutex1") != std::string::npos && 
		line.find("mutex1") != line.find(mutex1_str.str().c_str()) &&
		line.find("update_mutex") == std::string::npos &&
		line.find(update_mutex_str.str().c_str()) == std::string::npos)
	  search_str = "mutex1";
	else if(line.find("buffer_initialized") != std::string::npos && 
		line.find("buffer_initialized") != line.find(buffer_initialized_str.str().c_str()))
	  search_str = "buffer_initialized";
	else if(line.find("update_trig_buf") != std::string::npos && 
		line.find("update_trig_buf") != line.find(update_trig_buf_str.str().c_str()))
	  search_str = "update_trig_buf";
	else if(line.find("update_mutex") != std::string::npos && 
		line.find("update_mutex") != line.find(update_mutex_str.str().c_str()))
	  search_str = "update_mutex";
	else if(line.find("bh") != std::string::npos && 
		line.find("bh") != line.find(bh_str.str().c_str()))
	  search_str = "bh";
	else if(line.find("vec_buf") != std::string::npos && 
		line.find("vec_buf") != line.find(vec_buf_str.str().c_str()))
	  search_str = "vec_buf";
	else if(line.find("update_trig_bin") != std::string::npos &&
		line.find("update_trig_bin") != line.find(update_trig_bin_str.str().c_str()))
	  search_str = "update_trig_bin";
	else if(line.find("update_trig") != std::string::npos && 
		line.find("update_trig") != line.find(update_trig_str.str().c_str()) && 
		line.find("update_trig") != line.find(update_trig_bin_str.str().c_str()))
	  search_str = "update_trig";
	
	std::ostringstream search_str_str;
	search_str_str << search_str << view_id;
	if(line.find(search_str.c_str()) == line.find(search_str_str.str().c_str()))
	  search_str = "";
	
	if(search_str.length() != 0) {
	  size_t search_index = line.find(search_str.c_str());
	  std::ostringstream new_str;
	  new_str << line.substr(0, search_index) << search_str << view_id << line.substr(search_index + search_str.length());
	  line = new_str.str();
	  line_changed = true;
	}
      } while(line_changed);
    
      new_file << line << std::endl;
    }
    original_file.close();
    new_file.close();
  }
  else {
    std::cerr << "unable to open update_trig.c file" << std::endl;
    return -1;
  }

  //compile and install new original_update_trig_file
  //read postgres_path
  std::string postgres_path = "";

  std::ifstream hazy_conf_file(hazy_conf_file_path.c_str());
  if(hazy_conf_file.is_open()) {
    while(hazy_conf_file.good()) {
      getline(hazy_conf_file, line);
      if(line.find(POSTGRES_PATH_LINE_IN_CONF_FILE) != std::string::npos) {
	size_t eq_found = line.find_first_of("=");
	postgres_path = line.substr(eq_found + 1);
      }
    }
  }

  std::string s_libs_triggers_path = src_dir_path + "/" + INSTALLATION_DIR + "/" + S_LIBS_DIR + "/" + TRIGGERS_DIR;

  std::ostringstream compile_new_trigger;
  compile_new_trigger << GCC << " -I " << postgres_path << "/" << POSTGRES_INCLUDE << "/" 
		      << POSTGRES_SERVER << "  -I " << src_dir_path << "/" << INCLUDE_DIR 
		      << "/ -I " << src_dir_path << "/" << COMM_SERVER_DIR << "/ -I " 
		      << src_dir_path << "/" << SVM_CODE_DIR 
		      << "/ -D__POSTGRES_CALL -fPIC -o " << s_libs_triggers_path << "/" 
		      << UPDATE_TRIG_FILE_NAME << view_id << "." << OBJECT_FILE_EXT 
		      << " -c " << new_update_trig_file_name.str();

  std::ostringstream create_new_shared_file;
  create_new_shared_file << GPP << " -g -O3 -shared -W1,-soname," << UPDATE_TRIG_FILE_NAME 
			 << view_id << "." << SHARED_OBJECT_FILE_EXT << " -o " 
			 << s_libs_triggers_path << "/" << UPDATE_TRIG_FILE_NAME << view_id 
			 << "." << SHARED_OBJECT_FILE_EXT << " -L" << postgres_path 
			 << "/" << POSTGRES_LIB << " " << s_libs_triggers_path << "/" 
			 << UPDATE_TRIG_FILE_NAME << view_id << "." << OBJECT_FILE_EXT 
			 << " " << trigger_dir_path << "/low_level_net.o " 
			 << trigger_dir_path << "/hazy_message.o -lpq -lz -lm";


  //compile new trigger
  system(compile_new_trigger.str().c_str());
  system(create_new_shared_file.str().c_str());

  std::ostringstream create_update_trig_fnc_query;
  create_update_trig_fnc_query << "DROP FUNCTION IF EXISTS " << UPDATE_TRIG_FNC 
			       << view_id << "() CASCADE; CREATE OR REPLACE FUNCTION " 
			       << UPDATE_TRIG_FNC << view_id << "() RETURNS trigger AS '" 
			       << s_libs_triggers_path << "/" << UPDATE_TRIG_FILE_NAME 
			       << view_id << "." << SHARED_OBJECT_FILE_EXT << "' language c;";

  std::ostringstream create_update_trig_bin_fnc_query;
  create_update_trig_bin_fnc_query << "DROP FUNCTION IF EXISTS " << UPDATE_TRIG_BIN_FNC
			       << view_id << "() CASCADE; CREATE OR REPLACE FUNCTION " 
			       << UPDATE_TRIG_BIN_FNC << view_id << "() RETURNS trigger AS '" 
			       << s_libs_triggers_path << "/" << UPDATE_TRIG_FILE_NAME 
			       << view_id << "." << SHARED_OBJECT_FILE_EXT << "' language c;";

  std::ostringstream create_update_trig_on_example_table;
  create_update_trig_on_example_table << "DROP TRIGGER IF EXISTS " << UPDATE_TRIG_FNC_ON_TABLE << view_id << " ON " << this->example_table << "; CREATE TRIGGER " << UPDATE_TRIG_FNC_ON_TABLE << view_id << " AFTER INSERT OR UPDATE OR DELETE ON " << this->example_table << " FOR EACH ROW EXECUTE PROCEDURE " << UPDATE_TRIG_BIN_FNC << view_id << "();";

  int ret_val = db_conn->execute_statement_msg(create_update_trig_fnc_query.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, create_update_trig_fnc_query.str().c_str());
 
  ret_val = db_conn->execute_statement_msg(create_update_trig_bin_fnc_query.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, create_update_trig_bin_fnc_query.str().c_str());
  
  ret_val = db_conn->execute_statement_msg(create_update_trig_on_example_table.str().c_str(), error_msg.c_str());
  checkQueryReturnValue(ret_val, create_update_trig_on_example_table.str().c_str());

  return 0;
}
