#ifndef __hazy_create_view_h
#define __hazy_create_view_h

#include "postgres_util.h"
#include "globals.h"
#include "low_level_net.h"
#include "hazy_model.h"
#include "hazy_message.h"
#include <string>
#include "server_conn.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include "parse_conf_file.h"

class Hazy_Create_View {
 private:
  Hazy_Database *db_conn;
  std::string hazy_db_name;

  std::string view_name;
  std::string view_key;
  std::string original_entity_table;
  std::string hazy_entity_table;
  std::string entity_table_key;
  std::string labels_table;
  std::string label_in_table;
  std::string example_table;
  std::string example_table_key;
  std::string example_table_label;
  std::string feature_function;
  std::string training_table_name;

  enum storage_manager_type sm;
  int dense;
  int useepsmap;
  hazy_model::strategy hazy_strategy;
  int reservoir_size;
  double resort_time;
  double tao_val;
  double alpha_ratio;
  int buffer_size;
  int dim;
  double m_factor;

  int write_func_create_view(message_data msg, size_t length);
  std::string create_one_param_func_str(int view_id, 
				std::string fnc_name, 
				std::string param, 
				std::string return_type, 
				std::string so_file_source_base_name, std::string source_fnc_name);

 public:
  Hazy_Create_View(std::string, std::string, std::string, std::string, std::string, std::string, 
		   std::string, std::string, std::string, std::string, std::string, 
		   enum storage_manager_type, int, int, hazy_model::strategy, int, double, double,
		   double, int);
  void retrieve_unique_view_id(int &view_id);
  void insert_newview_to_catalog(int view_id);
  void callGetStatsFeaturesFunction(int view_id);
  void normalizeAndGetM(int view_id, double &m_factor);
  void callCreateEntityFeaturesFunction(int view_id);
  void callPrepareTrainFunction(int view_id);
  void getNumberOfFeatures(int view_id, int &num_of_features);
  int createBackendFunctions(int view_id);
  int createUpdateInPlaceFunction(int view_id);
  void createView(int view_id);
  int createTriggers(int view_id);
};

#endif
