#include <iostream>
#include <string>
#include "create_view.h"
#include "parse_conf_file.h"

using namespace std;

/*
 * This class handles view creation and creates all required functions & triggers
 */



int main(int argc, char *argv[]) {
  //check arg num
  if(argc != 11) {
    cerr << "wrong # of args" << endl;
    cerr << "exiting..." << endl;
    exit(1);
  }

  string view_name = argv[1];
  string view_key = argv[2];
  string entity_table = argv[3];
  string entity_table_key = argv[4];
  string labels_table = argv[5];
  string label_in_table = argv[6];
  string example_table = argv[7];
  string example_table_key = argv[8];
  string example_table_label = argv[9];
  string feature_function = argv[10];
  
  //retrieve dbname from conf file
  string hazy_conf_file_path = retrieve_hazy_conf_file_path();

  string dbname = "";
  int ret_val = parse_dbname_from_conf_file(hazy_conf_file_path, dbname);

  if(dbname.length() == 0) {
    cerr << "database name could not be parsed from hazy.conf" << endl;
    cerr << "exiting..." << endl;
    exit(1);
  }

  if(ret_val != SUCCESSFUL_PARSE) {
    handle_success_code(ret_val);
    exit(1);
  }

  double resort_time = 1.8;
  double tao_val = 0.001;
  double alpha_ratio = 1;
  int dense = 0;

  //Hazy parameters are retrieved from conf file
  enum storage_manager_type sm;
  int useepsmap;
  hazy_model::strategy hazy_strategy;
  int reservoir_size;
  int buffer_size;

  ret_val = parse_hazy_settings(hazy_conf_file_path, sm, useepsmap, hazy_strategy, reservoir_size, buffer_size);
  if(ret_val != SUCCESSFUL_PARSE) {
    handle_success_code(ret_val);
    exit(1);
  }

  
  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
						buffer_size);

  int view_id = -1;
  view->retrieve_unique_view_id(view_id);
 
  if(view_id == -1) {
    cerr << "The view could not be created, because unique view id retrieved is -1" << endl;
    exit(1);
  }
  else {
    //feature vectors are already processed?
    if(feature_function.compare("") == 0) {
      
    }
    //we should use tf idf?
    else {
      //Insert this new view id info to the catalog
      view->insert_newview_to_catalog(view_id);
      //Normalize vectors and retrieve m factor
      view->callGetStatsFeaturesFunction(view_id);
      double m_factor = 0.0;
      view->normalizeAndGetM(view_id, m_factor);
      view->callCreateEntityFeaturesFunction(view_id);
      view->callPrepareTrainFunction(view_id);
      int num_of_features = 0;
      view->getNumberOfFeatures(view_id, num_of_features);
      view->createView(view_id);
      view->createTriggers(view_id);
      cout << "view is successfully created with view id = " << view_id << endl;
    }
  }
}
