#include <iostream>
#include <string>
#include "create_view.h"

using namespace std;

string view_name = "dblife_labeled_papers";
string view_key = "id";
string entity_table = "dblife_pubs";
string entity_table_key = "id";
string labels_table = "Paper_Area";
string label_in_table = "l";
string example_table = "Example_Papers";
string example_table_key = "id";
string example_table_label = "l";
string feature_function = "tf_bag_of_words";
int global_counter = 0;

string dbname = "VLDB11_Exp";
//Also, these can be retrieved from conf file?
enum storage_manager_type sm = ON_DISK_SM;
int dense = 0;
int useepsmap = 0;
hazy_model::strategy hazy_strategy = hazy_model::EAGER_HAZY;
int reservoir_size = 1000;
double resort_time = 1.8;
double tao_val = 0.0001;
double alpha_ratio = 1;
int buffer_size = 0;

const std::string error_msg = "There is an error in the query\n";

struct _id_entry {
  int id;
  static int parse_tuple(PGresult *res, int index, _id_entry &c) {
    c.id = atoi(PQgetvalue(res, index, 0));
    return 0;
  }
};


void parse_pair(const char *str, int &feature_id, float &feature_value) {
  if (sscanf(str, " ( %d , %f )", &feature_id, &feature_value) != 2)
    cout << "there is a problem with the input: " << str << endl;
}

bool check_table_exist(Hazy_Database *db_conn, string table_name) {
  ostringstream check_query;
  check_query << "SELECT COUNT(*) FROM " << PG_CLASS_TABLE << " WHERE " << PG_CLASS_RELNAME_COL << " = '" << table_name
	      << "';";
  int count = 0;
  int ret_val = db_conn->execute_query_msg_int(check_query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, check_query.str().c_str());
  if(count != 1)
    return false;
  else
    return true;
}

void test_retrieve_unique_id(Hazy_Database *db_conn, bool &success) {
  success = true;

  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
						buffer_size);

  ostringstream query;
  query << "SELECT COUNT(*) FROM " << CLASSIFICATION_GENERAL_TABLE << ";";
  int count = 0;
  int ret_val = db_conn->execute_query_msg_int(query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, query.str().c_str());

  int view_id = -1;
  if(count == 0)
    view_id = 0;
  else {
    ostringstream query;
    query << "SELECT MAX(" << VIEW_ID << ") FROM " << CLASSIFICATION_GENERAL_TABLE << ";";
    int ret_val = db_conn->execute_query_msg_int(query.str().c_str(), error_msg.c_str(), view_id);
    checkQueryReturnValue(ret_val, query.str().c_str());
    view_id ++;
  }

  int retrieved_view_id = 0;
  view->retrieve_unique_view_id(retrieved_view_id);

  if(retrieved_view_id != view_id)
    success = false;
}

void test_insert_newview_to_catalog(Hazy_Database *db_conn, bool &success) {
  success = true;

  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
						buffer_size);

  int view_id = 0;
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  ostringstream check_query;
  check_query << "SELECT COUNT(*) FROM " << CLASSIFICATION_GENERAL_TABLE << " WHERE " << VIEW_ID << " = " << view_id << " AND "
	      << VIEW_NAME << " = '" << view_name << "' AND " << ENTITY_TABLE_NAME << " = '" << entity_table << "' AND "
	      << LABELS_TABLE_NAME << " = '" << labels_table << "' AND " << EXAMPLE_TABLE_NAME << " = '" << example_table << "' "
	      << " AND " << FEATURE_FUNCTION << " = '" << feature_function << "';";
  int count = 0;
  int ret_val = db_conn->execute_query_msg_int(check_query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, check_query.str().c_str());
  if(count != 1)
    success = false;
}

void test_callGetStatsFeaturesFunction(Hazy_Database *db_conn, bool &success) {
  success = true;

  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
						buffer_size);
  
  int view_id = 0;
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  view->callGetStatsFeaturesFunction(view_id);
  
  //check table is actually created
  ostringstream check_table_name;
  check_table_name << EXTRACTED_FEATURES << view_id;

  if(!check_table_exist(db_conn, check_table_name.str())) {
    cout << "[test_callGetStatsFeaturesFunction] extractedfeatures table does not exist" << endl;
    success = false;
  }

  //also check whether all tfidf scores > 0
  ostringstream check_score_query;
  check_score_query << "SELECT COUNT(*) FROM " << EXTRACTED_FEATURES << view_id << " WHERE " 
		    << EXTRACTED_FEATURES_FEATURE_VAL_COL << " <= 0;";

  int count = 0;
  int ret_val = db_conn->execute_query_msg_int(check_score_query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, check_score_query.str().c_str());
  if(count > 0) {
    cout << "[test_callGetStatsFeaturesFunction] there are negative feature values" << endl;
    cout << "query: " << check_score_query.str().c_str() << endl;
    success = false;
  }  
}

void test_normalizeAndGetM(Hazy_Database *db_conn, bool &success) {
  success = true;

  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
						buffer_size);
  
  int view_id = 0;
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  view->callGetStatsFeaturesFunction(view_id);
  double m_factor = 0.0;
  view->normalizeAndGetM(view_id, m_factor);

  ostringstream check_score_query;
  check_score_query << "SELECT COUNT(*) FROM " << EXTRACTED_FEATURES << view_id << " WHERE " 
		    << EXTRACTED_FEATURES_FEATURE_VAL_COL << " <= 0 or " << EXTRACTED_FEATURES_FEATURE_VAL_COL << " > 1;";
  
  int count = 0;
  int ret_val = db_conn->execute_query_msg_int(check_score_query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, check_score_query.str().c_str());
  if(count > 0) {
    cout << "[test_normalizeAndGetM] there are " << count << " feature values <= 0 or > 1" << endl;
    cout << "query: " << check_score_query.str().c_str() << endl;
    success = false;
  }
  if(m_factor - 1 >= 1.0e-5) {
    cout << "[test_normalizeAndGetM] m_factor >= 1 + 1.0e-5";
    success = false;
  }
}

void test_callCreateEntityFeaturesFunction(Hazy_Database *db_conn, bool &success) {
  success = true;
  
  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
  						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
  						buffer_size);
  
  int view_id = 0;
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  view->callGetStatsFeaturesFunction(view_id);
  double m_factor = 0.0;
  view->normalizeAndGetM(view_id, m_factor);
  view->callCreateEntityFeaturesFunction(view_id);

  ostringstream check_entity_vectors_table_name;
  check_entity_vectors_table_name << HAZY_ENTITY_TABLE_BASE_NAME << view_id;
  
  if(!check_table_exist(db_conn, check_entity_vectors_table_name.str())) {
    cout << "[test_callCreateEntityFeaturesFunction] training table does not exist" << endl;
    success = false;
  }

  ostringstream check_feature_ids_table_name;
  check_feature_ids_table_name << FEATURE_IDS_TABLE_BASE_NAME << view_id;

  if(!check_table_exist(db_conn, check_feature_ids_table_name.str())) {
    cout << "[test_callCreateEntityFeaturesFunction] feature_ids table does not exist" << endl;
    success = false;
  }

  //test # of features are consistent
  ostringstream check_dims_query;
  check_dims_query << "SELECT COUNT(*) FROM (SELECT id, array_length(feature_vector, 1) AS length from " << 
    HAZY_ENTITY_TABLE_BASE_NAME << view_id << ") AS r_test_1" << view_id << ", (SELECT id, count(*) AS count FROM " << 
    EXTRACTED_FEATURES << view_id << " GROUP BY id) AS r_test_2" << view_id << " WHERE r_test_1" << view_id << 
    ".id = r_test_2" << view_id << ".id AND r_test_1" << view_id << ".length <> r_test_2" << view_id << ".count;";

  int count = 0;
  int ret_val = db_conn->execute_query_msg_int(check_dims_query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, check_dims_query.str().c_str());
  if(count > 0) {
    cout << "[test_callCreateEntityFeaturesFunction] there are " << count << " tuples in hazy entity table whose feature counts dont match with the ones in the extracted features table" << endl;
    cout << "query: " << check_dims_query.str().c_str() << endl;
    success = false;
  }

  ostringstream create_index_query;
  create_index_query << "DROP INDEX IF EXISTS hazy_table_index" << view_id << ";"
		     << "CREATE INDEX hazy_table_index" << view_id
		     << " ON " << HAZY_ENTITY_TABLE_BASE_NAME << view_id << " USING hash(id);";
  db_conn->execute_statement_msg(create_index_query.str().c_str(), error_msg.c_str());

  //test features in hazy entity table exist in extracted features and they are ordered by feature_id
  std::vector<_id_entry> _tuples;
  ostringstream retrieve_query;
  retrieve_query << "SELECT id FROM " << HAZY_ENTITY_TABLE_BASE_NAME << view_id << " ORDER BY id;";
  db_conn->postgresTupleParser<_id_entry>(retrieve_query.str().c_str(), _id_entry::parse_tuple,
							    _tuples);

  for(unsigned int i = 0; i < _tuples.size(); i ++) {
    int tuple_id = _tuples.at(i).id;
    cout << "tuple id: " << tuple_id << endl;

    ostringstream retrieve_dim_query;
    retrieve_dim_query << "SELECT array_length(feature_vector, 1) FROM " << HAZY_ENTITY_TABLE_BASE_NAME << view_id
		       << " WHERE id = " << tuple_id << ";";
    
    int dim = 0;
    
    ret_val = db_conn->execute_query_msg_int(retrieve_dim_query.str().c_str(), error_msg.c_str(), dim);
    checkQueryReturnValue(ret_val, retrieve_dim_query.str().c_str());
    
    int old_feature_id = -1;
    if(dim < 1) {
      cerr << "problem with dim in [test_callCreateEntityFeaturesFunction]" << endl;
      success = false;
      break;
    }
    
    for(int j = 1; j <= dim; j ++) {
      ostringstream retrieve_pair_query;
      retrieve_pair_query << "SELECT feature_vector[" << j << "] FROM " << HAZY_ENTITY_TABLE_BASE_NAME << view_id
  			  << " WHERE id = " << tuple_id << ";";
      string pair_str = "";
      ret_val = db_conn->execute_query_msg_text(retrieve_pair_query.str().c_str(), error_msg.c_str(), pair_str);
      checkQueryReturnValue(ret_val, retrieve_pair_query.str().c_str());
      int new_feature_id = -1;
      float feature_value;
      parse_pair(pair_str.c_str(), new_feature_id, feature_value);

      if(new_feature_id == -1) {
  	cout << "[test_callCreateEntityFeaturesFunction] feature could not be parsed" << endl;
	cout << "\tdim was: " << dim << endl;
	cout << "\tdim query: " << retrieve_dim_query.str() << endl;
	cout << "\tid: " << tuple_id << endl;
	cout << "\tretrieve pair query: " << retrieve_pair_query.str() << endl;
  	success = false;
  	break;
      }
      
      if(new_feature_id < old_feature_id) {
  	cout << "[test_callCreateEntityFeaturesFunction] feature ids are not in increasing order" << endl;
  	success = false;
  	break;
      }
      
      ostringstream check_features_valid_query;
      check_features_valid_query << "SELECT COUNT(*) FROM " << EXTRACTED_FEATURES << view_id << " e, " << 
  	FEATURE_IDS_TABLE_BASE_NAME << view_id << " f, " << HAZY_ENTITY_TABLE_BASE_NAME << view_id << 
  	" t where e.feature = f.feature AND e.id = t.id AND t.id = " << tuple_id << " AND f.feature_id = " << 
  	new_feature_id << " AND abs(e.feature_value - " << feature_value << ") <= 1.0e-5;";
      
      int check_count_of_features = 0;
      ret_val = db_conn->execute_query_msg_int(check_features_valid_query.str().c_str(), error_msg.c_str(), 
  					       check_count_of_features);
      checkQueryReturnValue(ret_val, check_features_valid_query.str().c_str());
      
      if(check_count_of_features != 1) {
  	cout << "[test_callCreateEntityFeaturesFunction] (id, feature_id, feature_value) is either does not exist or exists more than one" << endl;
  	success = false;
  	break;
      }
      
      old_feature_id = new_feature_id;
    }
    if(!success)
      break;
  }
}

void test_callPrepareTrainFunction(Hazy_Database *db_conn, bool &success) {
  success = true;
  
  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
						buffer_size);
  
  int view_id = 0;
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  view->callGetStatsFeaturesFunction(view_id);
  double m_factor = 0.0;
  view->normalizeAndGetM(view_id, m_factor);
  view->callCreateEntityFeaturesFunction(view_id);
  view->callPrepareTrainFunction(view_id);

  ostringstream check_training_table_name;
  check_training_table_name << TRAINING_TABLE << view_id;
  
  if(!check_table_exist(db_conn, check_training_table_name.str())) {
    cout << "[test_callPrepareTrainFunction] training table does not exist" << endl;
    success = false;
  }

  ostringstream check_feature_ids_table_name;
  check_feature_ids_table_name << FEATURE_IDS_TABLE_BASE_NAME << view_id;

  if(!check_table_exist(db_conn, check_feature_ids_table_name.str())) {
    cout << "[test_callPrepareTrainFunction] feature_ids table does not exist" << endl;
    success = false;
  }

  //check all labels are correctly placed in training table
  ostringstream check_labels_query;
  check_labels_query << "SELECT COUNT(*) FROM " << example_table << " e, " << TRAINING_TABLE << view_id 
		     << " t WHERE e." << example_table_key
		     << " = t.id and e." << example_table_label << " <> t.label;";

  int count = 0;
  int ret_val = db_conn->execute_query_msg_int(check_labels_query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, check_labels_query.str().c_str());
  if(count > 0) {
    cout << "[test_callPrepareTrainFunction] there are " << count << " tuples in training table whose labels dont match with the ones in the example table" << endl;
    cout << "query: " << check_labels_query.str().c_str() << endl;
    success = false;
  }

  ostringstream check_dims_query;
  check_dims_query << "SELECT COUNT(*) FROM (SELECT id, array_length(feature_vector, 1) AS length from " << 
    TRAINING_TABLE << view_id << ") AS r_test_1" << view_id << ", (SELECT id, count(*) AS count FROM " << 
    EXTRACTED_FEATURES << view_id << " GROUP BY id) AS r_test_2" << view_id << " WHERE r_test_1" << view_id << 
    ".id = r_test_2" << view_id << ".id AND r_test_1" << view_id << ".length <> r_test_2" << view_id << ".count;";

  count = 0;
  ret_val = db_conn->execute_query_msg_int(check_dims_query.str().c_str(), error_msg.c_str(), count);
  checkQueryReturnValue(ret_val, check_dims_query.str().c_str());
  if(count > 0) {
    cout << "[test_callPrepareTrainFunction] there are " << count << " tuples in training table whose feature counts dont match with the ones in the extracted features table" << endl;
    cout << "query: " << check_dims_query.str().c_str() << endl;
    success = false;
  }

  ostringstream create_index_query;
  create_index_query << "DROP INDEX IF EXISTS hazy_table_index" << view_id << ";"
		     << "CREATE INDEX hazy_table_index" << view_id
		     << " ON " << TRAINING_TABLE << view_id << " USING hash(id);";
  db_conn->execute_statement_msg(create_index_query.str().c_str(), error_msg.c_str());

  //test features in training table exist in extracted features and they are ordered by feature_id
  std::vector<_id_entry> _tuples;
  ostringstream retrieve_query;
  retrieve_query << "SELECT id FROM " << TRAINING_TABLE << view_id << " ORDER BY id;";
  db_conn->postgresTupleParser<_id_entry>(retrieve_query.str().c_str(), _id_entry::parse_tuple,
							    _tuples);

  for(unsigned int i = 0; i < _tuples.size(); i ++) {
    int tuple_id = _tuples.at(i).id;

    ostringstream retrieve_dim_query;
    retrieve_dim_query << "SELECT array_length(feature_vector, 1) FROM " << TRAINING_TABLE << view_id
		       << " WHERE id = " << tuple_id << ";";
    
    int dim = 0;
    
    ret_val = db_conn->execute_query_msg_int(retrieve_dim_query.str().c_str(), error_msg.c_str(), dim);
    checkQueryReturnValue(ret_val, retrieve_dim_query.str().c_str());
    
    int old_feature_id = -1;
    if(dim < 1) {
      cerr << "problem with dim in [test_callPrepareTrainFunction]" << endl;
      success = false;
      break;
    }
    
    for(int j = 1; j <= dim; j ++) {
      ostringstream retrieve_pair_query;
      retrieve_pair_query << "SELECT feature_vector[" << j << "] FROM " << TRAINING_TABLE << view_id
  			  << " WHERE id = " << tuple_id << ";";
      string pair_str = "";
      ret_val = db_conn->execute_query_msg_text(retrieve_pair_query.str().c_str(), error_msg.c_str(), pair_str);
      checkQueryReturnValue(ret_val, retrieve_pair_query.str().c_str());
      int new_feature_id = -1;
      float feature_value;
      parse_pair(pair_str.c_str(), new_feature_id, feature_value);

      if(new_feature_id == -1) {
  	cout << "[test_callPrepareTrainFunction] feature could not be parsed" << endl;
	cout << "\tdim was: " << dim << endl;
	cout << "\tdim query: " << retrieve_dim_query.str() << endl;
	cout << "\tid: " << tuple_id << endl;
	cout << "\tretrieve pair query: " << retrieve_pair_query.str() << endl;
  	success = false;
  	break;
      }
      
      if(new_feature_id < old_feature_id) {
  	cout << "[test_callPrepareTrainFunction] feature ids are not in increasing order" << endl;
  	success = false;
  	break;
      }
      
      ostringstream check_features_valid_query;
      check_features_valid_query << "SELECT COUNT(*) FROM " << EXTRACTED_FEATURES << view_id << " e, " << 
  	FEATURE_IDS_TABLE_BASE_NAME << view_id << " f, " << TRAINING_TABLE << view_id << 
  	" t where e.feature = f.feature AND e.id = t.id AND t.id = " << tuple_id << " AND f.feature_id = " << 
  	new_feature_id << " AND abs(e.feature_value - " << feature_value << ") <= 1.0e-5;";
      
      int check_count_of_features = 0;
      ret_val = db_conn->execute_query_msg_int(check_features_valid_query.str().c_str(), error_msg.c_str(), 
  					       check_count_of_features);
      checkQueryReturnValue(ret_val, check_features_valid_query.str().c_str());
      
      if(check_count_of_features != 1) {
  	cout << "[test_callPrepareTrainFunction] (id, feature_id, feature_value) is either does not exist or exists more than one" << endl;
  	success = false;
  	break;
      }
      
      old_feature_id = new_feature_id;
    }
    if(!success)
      break;
  }
}

void test_getNumberOfFeatures(Hazy_Database *db_conn, bool &success) {
  success = true;

  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
						buffer_size);  
  
  int view_id = 0;
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  view->callGetStatsFeaturesFunction(view_id);
  double m_factor = 0.0;
  view->normalizeAndGetM(view_id, m_factor);
  view->callCreateEntityFeaturesFunction(view_id);
  view->callPrepareTrainFunction(view_id);
  int num_of_features = 0;
  view->getNumberOfFeatures(view_id, num_of_features);

  int check_num_of_features = 0;
  ostringstream check_num_of_features_query;
  check_num_of_features_query << "SELECT COUNT(*) FROM " << FEATURE_IDS_TABLE_BASE_NAME << view_id << ";";
  int ret_val = db_conn->execute_query_msg_int(check_num_of_features_query.str().c_str(), error_msg.c_str(), 
					       check_num_of_features);
  checkQueryReturnValue(ret_val, check_num_of_features_query.str().c_str());
  if(num_of_features != check_num_of_features) {
    cout << "[test_getNumberOfFeatures] num of features retrieved is not correct" << endl;
    success = false;
  }
}

void test_singleCreateView(Hazy_Database *db_conn, bool &success, enum storage_manager_type _sm, int _useepsmap, 
			   hazy_model::strategy _strategy, int _buffer_size) {
  ostringstream new_view_name;
  new_view_name << view_name << global_counter;
  global_counter ++;
  sm = _sm;
  dense = 0;
  useepsmap = _useepsmap;
  hazy_strategy = _strategy;
  reservoir_size = 1000;
  resort_time = 1.8;
  tao_val = 0.0001;
  alpha_ratio = 1;
  buffer_size = _buffer_size;
  view_name = new_view_name.str();

  success = true;

  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
						buffer_size);  
  
  int view_id = 0;
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  view->callGetStatsFeaturesFunction(view_id);
  double m_factor = 0.0;
  view->normalizeAndGetM(view_id, m_factor);
  view->callCreateEntityFeaturesFunction(view_id);
  view->callPrepareTrainFunction(view_id);
  int num_of_features = 0;
  view->getNumberOfFeatures(view_id, num_of_features);
  view->createView(view_id);
  if(sm == ON_DISK_SM || sm == HYBRID_SM) {
    //check internal and view tables exist
    if(hazy_strategy == hazy_model::LAZY_HAZY || hazy_strategy == hazy_model::EAGER_HAZY) {
      string internal_table_base_name = "internal_table";
      
      ostringstream internal_table_name;
      
      internal_table_name << internal_table_base_name << view_id;
      
      if(!check_table_exist(db_conn, internal_table_name.str())) {
	cout << "[test_singleCreateView] internal table does not exist: " << internal_table_name.str() << endl;
	success = false;
      }
    }

    if(!check_table_exist(db_conn, view_name)) {
      cout << "[test_singleCreateView] view table does not exist" << endl;
      success = false;
    }
  }
}

void test_createView(Hazy_Database *db_conn, bool &success) {
  test_singleCreateView(db_conn, success, MAIN_MEMORY_SM, 0, hazy_model::LAZY_NAIVE, 0);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, MAIN_MEMORY_SM, 0, hazy_model::EAGER_NAIVE, 0);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, MAIN_MEMORY_SM, 0, hazy_model::LAZY_HAZY, 0);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, MAIN_MEMORY_SM, 0, hazy_model::EAGER_HAZY, 0);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, ON_DISK_SM, 0, hazy_model::LAZY_NAIVE, 0);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, ON_DISK_SM, 0, hazy_model::EAGER_NAIVE, 0);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, ON_DISK_SM, 0, hazy_model::LAZY_HAZY, 0);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, ON_DISK_SM, 0, hazy_model::EAGER_HAZY, 0);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, HYBRID_SM, 0, hazy_model::LAZY_HAZY, 1239);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, HYBRID_SM, 1, hazy_model::LAZY_HAZY, 1239);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, HYBRID_SM, 0, hazy_model::EAGER_HAZY, 1239);
  if(!success)
    return;
  test_singleCreateView(db_conn, success, HYBRID_SM, 1, hazy_model::EAGER_HAZY, 1239);
  if(!success)
    return;
}

void test_createTriggers(Hazy_Database *db_conn, bool &success) {
  success = true;

  Hazy_Create_View *view = new Hazy_Create_View(dbname, view_name, view_key, entity_table, entity_table_key,
  						labels_table, label_in_table, example_table, example_table_key,
  						example_table_label, feature_function, sm, dense, useepsmap,
  						hazy_strategy, reservoir_size, resort_time, tao_val, alpha_ratio,
  						buffer_size);  
  
  int view_id = 0;
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  view->callGetStatsFeaturesFunction(view_id);
  double m_factor = 0.0;
  view->normalizeAndGetM(view_id, m_factor);
  view->callCreateEntityFeaturesFunction(view_id);
  view->callPrepareTrainFunction(view_id);
  int num_of_features = 0;
  view->getNumberOfFeatures(view_id, num_of_features);
  view->createView(view_id);
  view->createTriggers(view_id);
}


void create_view(string db_n, string vn, string vk, string et, string etk, 
			      string lt, string lit, string ext, string extk, 
		 string extl, string ff, int &view_id) {
  Hazy_Create_View *view = new Hazy_Create_View(db_n, vn, vk, et, etk, lt, lit, ext, extk, extl, ff, 
						sm, dense, useepsmap, hazy_strategy, reservoir_size, 
						resort_time, tao_val, alpha_ratio, buffer_size);  
  
  view->retrieve_unique_view_id(view_id);
  view->insert_newview_to_catalog(view_id);
  view->callGetStatsFeaturesFunction(view_id);
  double m_factor = 0.0;
  view->normalizeAndGetM(view_id, m_factor);
  view->callCreateEntityFeaturesFunction(view_id);
  view->callPrepareTrainFunction(view_id);
  int num_of_features = 0;
  view->getNumberOfFeatures(view_id, num_of_features);
  view->createView(view_id);
  view->createTriggers(view_id);
}


void insert_stmt(Hazy_Database *db_conn, string ex_t_name, int view_id, int ent_id, int ent_label) {
  ostringstream insert_query;
  insert_query << "INSERT INTO " << ex_t_name << " VALUES(" << ent_id << ", " << ent_label << ", " 
	       << "(SELECT feature_vector FROM " << "hazy_entity_table" << view_id 
	       << " WHERE id = " << ent_id << ")" << ")";

  int ret_val = db_conn->execute_statement_msg(insert_query.str().c_str(), "insert update example query");
  checkQueryReturnValue(ret_val, insert_query.str().c_str());
}

void test_handleMultipleViewUpdates(bool &success) {
  success = true;

  string db_n = "hazy_demo_database";
  Hazy_Database *db_conn = new Hazy_Database(db_n);

  //set view parameters
  string view_name1 = "reg_view1";
  string view_name2 = "reg_view2";
  string view_key = "id";
  string entity_table = "demo_papers";
  string entity_table_key = "id";
  string labels_table = "Paper_Area";
  string label_in_table = "l";
  string example_table1 = "regression_example1";
  string example_table2 = "regression_example2";
  string example_table_key = "id";
  string example_table_label = "l";
  string feature_function = "tf_bag_of_words";

  //set reservoir size 0 to test instantly
  reservoir_size = 0;

  //create training_example1
  string create_tr_exs = "DROP TABLE IF EXISTS regression_example1 CASCADE;";
  create_tr_exs += "CREATE TABLE regression_example1(id int, l int, feature_vector pair[]);";
  create_tr_exs += "DROP TABLE IF EXISTS regression_example2 CASCADE;";
  create_tr_exs += "CREATE TABLE regression_example2(id int, l int, feature_vector pair[]);";
  create_tr_exs += "UPDATE hazy_cls_view_catalog SET view_name='reg_view11' WHERE view_name = 'reg_view1';";
  create_tr_exs += "UPDATE hazy_cls_view_catalog SET view_name='reg_view12' WHERE view_name = 'reg_view2';";

  int ret_val = db_conn->execute_statement_msg(create_tr_exs.c_str(), "create example table query");
  checkQueryReturnValue(ret_val, create_tr_exs.c_str());

  string drop_prev_view = "DROP TABLE IF EXISTS " + view_name1 + " CASCADE;";
  drop_prev_view += "DROP VIEW IF EXISTS " + view_name1 + " CASCADE;";
  drop_prev_view += "DROP TABLE IF EXISTS " + view_name2 + " CASCADE;";
  drop_prev_view += "DROP VIEW IF EXISTS " + view_name2 + " CASCADE;";

  ret_val = db_conn->execute_statement_msg(drop_prev_view.c_str(), "create example table query");
  checkQueryReturnValue(ret_val, drop_prev_view.c_str());

  int view_id1 = 0;
  int view_id2 = 0;

  create_view(db_n, view_name1, view_key, entity_table, entity_table_key,
	      labels_table, label_in_table, example_table1, example_table_key,
	      example_table_label, feature_function, view_id1);

  insert_stmt(db_conn, example_table1, view_id1, 5, 1);

  create_view(db_n, view_name2, view_key, entity_table, entity_table_key,
  	      labels_table, label_in_table, example_table2, example_table_key,
  	      example_table_label, feature_function, view_id2);
  
  // ostringstream test_query1;
  // test_query1 << "SELECT initialize_waste_count(); SELECT cache_model_waste('{0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,1.1920885891243957e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,2.1994803502213984e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,-7.3931187719757741e-01,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,2.1994803502213984e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,-1.2560687032080420e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,2.1994803502213984e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,2.1994803502213984e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,-1.2560687032080420e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,-1.0335136520209192e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,-1.2560687032080420e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,-1.2560687032080420e+00,-1.2560687032080420e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,-9.0332729273038048e-01,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,-1.0335136520209192e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00,0.0000000000000000e+00}'::float8[], 9.9900099900115946e-05);";

  // ret_val = db_conn->execute_statement_msg(test_query1.str().c_str(), "test1");
  // checkQueryReturnValue(ret_val, test_query1.str().c_str());

  // ostringstream test_query2;
  // test_query2 << "SELECT update_external_table_inplace" << view_id1 << "(e, dotprdct_cached_waste(i.feature_vector) > 0) FROM internal_table" << view_id1 << " i, reg_view1 e WHERE (i.eps BETWEEN -2.30168 AND 2.19938) AND (e.eps BETWEEN -2.30168 AND 2.19938) AND (e.id = i.id);";
  // ret_val = db_conn->execute_statement_msg(test_query2.str().c_str(), "test2");
  // checkQueryReturnValue(ret_val, test_query2.str().c_str());

  insert_stmt(db_conn, example_table1, view_id1, 35, -1);
  // insert_stmt(db_conn, example_table1, view_id1, 8, 1);
  // insert_stmt(db_conn, example_table1, view_id1, 42, -1);
  // insert_stmt(db_conn, example_table1, view_id1, 2, 1);
  // insert_stmt(db_conn, example_table2, view_id2, 1, 1);
  // insert_stmt(db_conn, example_table1, view_id1, 49, -1);
}

int main(int argc, char *argv[]) {
  Hazy_Database *db_conn = new Hazy_Database(dbname);
  bool success = false;

  //create sample example table
  string create_example_table_query = "DROP TABLE IF EXISTS example_papers;";
  create_example_table_query += "CREATE TABLE example_papers(id int, l int);";
  create_example_table_query += "INSERT INTO example_papers VALUES(70036, 1), (70648, -1), (70922, -1), (71242, 1), (71300, 1), (71625, -1), (71685, -1);";

  int ret_val = db_conn->execute_statement_msg(create_example_table_query.c_str(), "create example table query");
  checkQueryReturnValue(ret_val, create_example_table_query.c_str());

  // test_retrieve_unique_id(db_conn, success);
  // cout << "[test_retrieve_unique_id]: " << (success ? "success" : "failure") << endl;
  // test_insert_newview_to_catalog(db_conn, success);
  // cout << "[test_insert_newview_to_catalog]: " << (success ? "success" : "failure") << endl;
  // test_callGetStatsFeaturesFunction(db_conn, success);
  // cout << "[test_callGetStatsFeaturesFunction]: " << (success ? "success" : "failure") << endl;
  // test_normalizeAndGetM(db_conn, success);
  // cout << "[test_normalizeAndGetM]: " << (success ? "success" : "failure") << endl;
  // test_callCreateEntityFeaturesFunction(db_conn, success);
  // cout << "[test_callCreateEntityFeaturesFunction]: " << (success ? "success" : "failure") << endl;
  // test_callPrepareTrainFunction(db_conn, success);
  // cout << "[test_callPrepareTrainFunction]: " << (success ? "success" : "failure") << endl;
  // test_getNumberOfFeatures(db_conn, success);
  // cout << "[test_getNumberOfFeatures]: " << (success ? "success" : "failure") << endl;
  // test_createView(db_conn, success);
  // cout << "[test_createView]: " << (success ? "success" : "failure") << endl;
  // test_createTriggers(db_conn, success);
  // cout << "[test_createTriggers]: " << (success ? "success" : "failure") << endl;
  test_handleMultipleViewUpdates(success);
  cout << "[test_handleMultipleViewUpdates]: " << (success ? "success" : "failure") << endl;

  return 0;
}
