#include "hazy_message.h"
#include "postgres_util.h"
#include "server_conn.h"
#include "hashing.h"
#include "hazy_sgd.h"
#include "globals.h"
#include "frontend_util.h"
#include <errno.h>
#include <string>
#include <pthread.h>

#define HAZY_HYBRID "hazy_hybrid"
#define HAZY_ON_DISK "hazy_on_disk"
#define HAZY_MM "hazy_mm"

const string EXTERNAL_TABLE_BASE_NAME = "sm_test_external_table";
const string INTERNAL_TABLE_BASE_NAME = "internal_table";

void usage() {
  cerr << "usage: " << endl;
  return;
}

/**
 * This class is implemented for read experiments
 */

using namespace std;

///Hazy Server
Hazy_Server *serv;

template<class T>
void startServer(int server_id, int strategy_algorithm, int res_opt, int dim, double initial_resort, string dbname, double tao_val, double m_factor, double ratio, string entity_storage_name, int dense_or_sparse, bool mm,
		 int hybrid, int buffer_size, int useepsmap, model &m1, model &m2, bool fromSnapshot, model &initial_model, bool use_initial_model, double low_water, double high_water);

int main(int argc, char **argv) {
  int mm_hazy_version = 0; //default
  int strategy_algorithm = 0; //default
  int resOpt = 1000; //default
  int dim = 0;	//req
  double initial_resort = 0; //default
  string dbname = "";	//req
  double taoVal = 0.001; //default
  double m_factor = 0;//req
  double ratio = 1;	//optional
  string entity_table_name = ""; //req
  int dense_feature = 0;	//default
  int server_id = 0;	//req
  string model1_filepath = "";
  string model2_filepath = "";
  int hybrid = 0;
  int useepsmap;
  int buffer_size = 0;
  string initial_model_filepath = "";
  double low_water = 0;
  double high_water = 0;
  
  int c = 0, option_index = 0;
  char* usage_str = (char *) " <dim> <dbname> <m factor> <entity table name> <server id>";
  opterr = 1;
  
  static struct extended_option long_options[] = {
    {"mainmemory"  , no_argument, &mm_hazy_version, 1, "if want to use main memory (instead of on-disk)"},
    {"hybrid"  , no_argument, &hybrid, 1, "if want to use hybrid (instead of on-disk)"},
    {"dense"    , no_argument, &dense_feature, 1, "if input contains dense vectors"},
    {"useepsmap"    , no_argument, &useepsmap, 1, "use epsmap in hybrid"},
    {"strategy", required_argument, NULL, 's', "strategy:0:LAZY_NAIVE (default),1: EAGER_NAIVE,2:LAZY_HAZY,3:EAGER_HAZY"},
    {"reservoir", required_argument, NULL, 'r', "size of resevoir to be used by ResevoirSGD (default: 1000)"},
    {"resort_time", required_argument, NULL, 'i', "initial resort time (default 0)"},
    {"tao_val", required_argument, NULL, 't', "tao value in algorithm (default 0.001)"},
    {"alpha_ratio", required_argument, NULL, 'a', "alpha ratio in the paper (default 1)"},
    {"buf_size", required_argument, NULL, 'b', "hybrid buffer size"},
    {"initial_model", required_argument, NULL, 'l', "initial model"},
    {"model1", required_argument, NULL, 'm', "filename of the model that internal table sorted"},
    {"model2", required_argument, NULL, 'n', "filename of the model that is updated"},
    {"low_water", required_argument, NULL, 'g', "filename of the model that is updated"},
    {"high_water", required_argument, NULL, 'h', "filename of the model that is updated"},
    {NULL,0,NULL,0,0}
  };
  
  // Somehow this parsing is not working as advertised.
  struct option* lo = convert_extended_options(long_options);
  while( (c = getopt_long(argc, argv, "r:d:m:o:l:", lo, &option_index)) != -1)
    {
      switch (c) {
      case 's':
	strategy_algorithm = atoi(optarg);
	break;
      case 'r':
        resOpt = atoi(optarg);
        break;
      case 'i':
	initial_resort = atof(optarg);
	break;
      case 't':
	taoVal = atof(optarg);
	break;
      case 'a':
	ratio = atof(optarg);
	break;
      case 'm':
	model1_filepath = optarg;
	break;
      case 'n':
	model2_filepath = optarg;
	break;
      case 'b':
	buffer_size = atoi(optarg);
	break;
      case 'g':
	low_water = atof(optarg);
	break;
      case 'h':
	high_water = atof(optarg);
	break;
      case 'l':
	initial_model_filepath = optarg;
	break;
      case ':':
      case '?':
        print_usage(long_options, argv[0], usage_str);
      exit(-1);
      break;
      }
    }
  
  if(optind == argc - 5) {
    dim = atoi(argv[optind]);
    dbname = argv[optind + 1];
    m_factor = atof(argv[optind + 2]);
    entity_table_name = argv[optind + 3];
    server_id = atoi(argv[optind + 4]);
  } else {
    print_usage(long_options, argv[0], usage_str);
    exit(-1);
  }
  
  if((model1_filepath.compare("") == 0) ^ (model2_filepath.compare("") == 0)) {
    print_usage(long_options, argv[0], usage_str);
    exit(-1);
  }
  
  /*	///hazy version: HAZY_HYBRID, HAZY_DISK or HAZY_MM
	int mm_hazy_version = atoi(argv[1]);
	///strategy is eager or lazy and algorithm is hazy or naive
	int strategy_algorithm = atoi(argv[2]);
	///reservoir option: reservoir size (0 or more)
	int resOpt = atoi(argv[3]);
	///dimension of dataset (these are for experiments. for real system, it will automatically be written)
	int dim = atoi(argv[4]);
	///initial resort time
	double initial_resort = atof(argv[5]);
	///database name
	string dbname = argv[6];
	///tao value
	double taoVal = atof(argv[7]);
	///m_factor (this is for experiment. For real system, it will automatically be determined)
	double m_factor = atof(argv[8]);
	///this was for alpha sensitivity experiment
	double ratio = atof(argv[9]);
	
	///mm features
	string entity_table_name = argv[10];
	
	///on-disk & hybrid features
	int dense_feature = atoi(argv[11]);
	
	int server_id = atoi(argv[12]);*/
  
  bool fromSnapshot = false;
  bool use_initial_model = false;
  model m1, m2;
  model initial_model;
  
  if(model1_filepath.compare("") == 0) {
    m1 = 0;
    m2 = 0;
  }
  else {
    fromSnapshot = true;
    std::ifstream f1(model1_filepath.c_str());
    f1 >> m1;
    std::ifstream f2(model2_filepath.c_str());
    f2 >> m2;
  }
  
  if(initial_model_filepath.compare("") == 0) {
    initial_model = 0;
  }
  else {
    use_initial_model = true;
    std::ifstream f(initial_model_filepath.c_str());
    f >> initial_model;
  }
  
  LOGGING_ONLY(cout << "before function call, in hazy main m1: " << m1.w << endl; cout << "in hazy main m2: " << m2.w << endl;);
  if(dense_feature)
    startServer<FVector>(server_id, strategy_algorithm, resOpt, dim, initial_resort, dbname, taoVal, m_factor, ratio, entity_table_name, dense_feature, mm_hazy_version, hybrid, buffer_size, useepsmap, m1, m2, fromSnapshot,
			 initial_model, use_initial_model, low_water, high_water);
  else
    startServer<SVector>(server_id, strategy_algorithm, resOpt, dim, initial_resort, dbname, taoVal, m_factor, ratio, entity_table_name, dense_feature, mm_hazy_version, hybrid, buffer_size, useepsmap, m1, m2, fromSnapshot,
			 initial_model, use_initial_model, low_water, high_water);
  
  return 0;
}

template<class T>
struct worker_thread_info {
  Hazy_Worker *hw;
  Read_Specialized_Hazy_Sgd<T> *sgd;
  int server_id;
  
  worker_thread_info(Hazy_Worker *_hw, Read_Specialized_Hazy_Sgd<T> *_sgd, int _server_id) {
    hw = _hw;
    sgd = _sgd;
    server_id = _server_id;
  }
};

template<class T>
void*
worker_thread(void *data) {
  worker_thread_info<T> *wit = (worker_thread_info<T> *) data;
  Hazy_Worker *hw = wit->hw;
  Read_Specialized_Hazy_Sgd<T> *sgd = wit->sgd;
  int server_id = wit->server_id;
  
  int nSize = 2048;
  message_data answer = (char *) malloc(nSize * sizeof(char));
  bool close_conn = false;
  
  while (1) {
    void *data = hw->readData();
    if(data == NULL) {
      break;
    }
    
    switch(get_message_type(data)) {
    case READ_SE_LAZY_NAIVE:
      {
	int view_id = 0;
	T vec;
	parse_se_lazy_naive_msg(data, view_id, vec);
	int entity_label = -1;
	sgd->readEntityClass(vec, entity_label);
	LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
	if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	  cerr << "problem in build msg" << endl;
	  return NULL;
	}
      }
      break;
    case READ_SE_LAZY_NAIVE_BIN:
      {
	int view_id = 0;
        T vec;
        parse_se_lazy_naive_msg_bin(data, view_id, vec);
        int entity_label = -1;
        sgd->readEntityClass(vec, entity_label);
        LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
        if(build_read_se_answer(entity_label, answer, nSize) < 0) {
          cerr << "problem in build msg" << endl;
          return NULL;
        }
      }
      break;
    case READ_SE_LAZY_HAZY:
      {
	int view_id = 0;
	T vec;
	double eps = 0;
	parse_se_lazy_hazy_msg(data, view_id, vec, eps);
	int entity_label = -1;
	sgd->readEntityClass(vec, eps, entity_label);
	LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
	if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	  cerr << "problem in build msg";
	  return NULL;
        }
      }
      break;
    case READ_SE_LAZY_HAZY_BIN:
      { 
        int view_id = 0;
        T vec;
        double eps = 0;
        parse_se_lazy_hazy_msg_bin(data, view_id, vec, eps);
        int entity_label = -1;
        sgd->readEntityClass(vec, eps, entity_label);
        LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
        if(build_read_se_answer(entity_label, answer, nSize) < 0) {
          cerr << "problem in build msg";
          return NULL;
        }
      } 
      break;
    case READ_SE_EAGER:
      {
	int view_id = 0;
	int entity_id = 0;
	parse_se_eager_msg(data, view_id, entity_id);
	int entity_label = -1;
	sgd->readEntityClass(entity_id, entity_label);
	LOGGING_ONLY(cout << "label: " << entity_label << endl;);
	
      	if(build_read_se_answer(entity_label, answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
      	}
      }
      break;
    case READ_COUNT:
      {
	int view_id = 0;
	parse_count_msg(data, view_id);
	int numInClass = 0;
	sgd->readNumInClass(1, numInClass);
	LOGGING_ONLY(cout << "count: " << numInClass << endl;);
	
      	if(build_read_count_answer(numInClass, answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
      	}
      }
      break;
    case UPDATE:
      {
	int view_id = 0;
	T vec;
	int classOfExample = 0;
	parse_update_msg(data, view_id, vec, classOfExample);
	sgd->updateModel(vec, classOfExample);
	
	if(build_update_answer(answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
      	}	
      }
      break;
    case UPDATE_BIN:
      {
        int view_id = 0;
        T vec;
        int classOfExample = 0;
        parse_update_msg_bin(data, view_id, vec, classOfExample);
        sgd->updateModel(vec, classOfExample);
	
        if(build_update_answer(answer, nSize) < 0) {
          cerr << "problem in build msg";
          return NULL;
        }
      }   
      break;
    case QUERY_SERVER_ID:
      {
	if(build_query_server_answer(server_id, answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
	}
      }
      break;
    case CLOSE_CONNECTION:
      {
	close_conn = true;
	if(build_close_connection_answer((int) close_conn, answer, nSize) < 0) {
	  cerr << "problem in build msg";
	  return NULL;
	}
      }
      break;
    case SERVER_NOOP:
      {
	int ack = 1;
	if(build_noop_server_answer(ack, answer, nSize) < 0) {
	  cerr << "problem in build msg";
          return NULL;
	}
      }
      break;
    default: 
      cerr << "there is no such function code (" << get_message_type(data) << ")" << endl;
    }
    
    int success = hw->writeData(answer, strlen((char *) answer), false);
    if(success < 0)
      cerr << "error when writing to the client: " << strerror(errno);
    if(close_conn)
      break;
  }
  
  ///Server's client connection is closed
  hw->closeClientConnection();
  delete hw;
  if(close_conn) {
    ///Server connection is closed
    cout << "server is being closed" << endl;
    serv->closeConnection();
    exit(0);
  }
  return NULL;
}


template<class T>
void startServer(int server_id, int strategy_algorithm, int res_opt, int dim, double initial_resort, string dbname, double tao_val, double m_factor, double ratio, string entity_table_name, int dense_or_sparse, bool mm,
		 int hybrid, int buffer_size, int useepsmap, model &m1, model &m2, bool fromSnapshot, model &initial_model, bool use_initial_model, double low_water, double high_water) {
  ///database connector
  Hazy_Database *db_conn = new Hazy_Database(dbname);
  vector<pthread_t> workers;
  
  //install prepared statement	//should be in setup of hazy_sgd
  
  string parsedWord;
  
  storage_spec *spec = mm ? ((storage_spec*) new main_memory_storage_spec()) : ((storage_spec*) new ondisk_storage_spec());
  //in the experiment, we use view_id as 0. Therefore, unique_id is also 0, but actually they dont need to be consistent.
  spec->unique_id         = 0;
  spec->entity_table_name = entity_table_name; //file name for mm?
  spec->bRebuild = true;
  switch(strategy_algorithm) {
  case 0: spec->strategy = hazy_model::LAZY_NAIVE;
    break;
  case 1: spec->strategy = hazy_model::EAGER_NAIVE;
    break;
  case 2: spec->strategy = hazy_model::LAZY_HAZY;
    break;
  case 3: spec->strategy = hazy_model::EAGER_HAZY;
    break;
  default: cerr << "problem with strategy & algorithm " << endl; usage();
  }
  if(!mm) {
    ((ondisk_storage_spec*) spec)->external_table_name = EXTERNAL_TABLE_BASE_NAME;
    ((ondisk_storage_spec*) spec)->intermediate_table_name = INTERNAL_TABLE_BASE_NAME;
  }
  
  Read_Specialized_Hazy_Sgd<T> *sgd = NULL;
  
  // TAKE FROM COMMAND LINE!
  if(hybrid == 1) {
    if(!fromSnapshot) {
      if(use_initial_model) {
	//incremental sgd & hazy model
	IncrementalSGD<T> *inc_sgd = (res_opt > 0) ? (IncrementalSGD<T> *) new ReservoirSGD<T>(dim, LAMBDA, res_opt)
	  : (IncrementalSGD<T> *) new TrueSGD<T>(dim, LAMBDA);
	
	//if the user wants to start with a given model, then we use that model
	if(use_initial_model)
	  inc_sgd->setModel(initial_model);
	
	hazy_model hm(inc_sgd->getModel(), spec->strategy);
	//storage manager
	Storage_Manager<T> *st_manager;
	Ondisk_Storage_Manager_Factory<T> x;
	Ondisk_Storage_Manager<T> *disk = x.build_storage_manager(db_conn, (ondisk_storage_spec*) spec, hm);
	st_manager = (Storage_Manager<T> *) new Hybrid_Storage_Manager<T>(db_conn, (ondisk_storage_spec*) spec, disk, 
									  buffer_size, useepsmap == 1);
	//skiing spec
	skiing_spec ski_spec;
	ski_spec.initial_resort = initial_resort;
	ski_spec.taoVal = tao_val;
	ski_spec.m_factor = m_factor;
	ski_spec.ratio = ratio;
	Skiing ski_obj(ski_spec);
	
	sgd = new Read_Specialized_Hazy_Sgd<T>(st_manager, inc_sgd, ski_obj, spec->strategy, low_water, high_water);
      }
    }
  }
  else {
    if(!fromSnapshot) {
      if(use_initial_model) {
	//incremental sgd & hazy model
	IncrementalSGD<T> *inc_sgd = (res_opt > 0) ? (IncrementalSGD<T> *) new ReservoirSGD<T>(dim, LAMBDA, res_opt)
	  : (IncrementalSGD<T> *) new TrueSGD<T>(dim, LAMBDA);
	
	//if the user wants to start with a given model, then we use that model
	if(use_initial_model)
	  inc_sgd->setModel(initial_model);
	
	hazy_model hm(inc_sgd->getModel(), spec->strategy);
	//storage manager
	Storage_Manager<T> *st_manager;
	Ondisk_Storage_Manager_Factory<T> x;
	st_manager = (mm) ? ((Storage_Manager<T> *) new STLMainMemory_Storage_Manager<T>(db_conn, 
											 (main_memory_storage_spec*) spec, hm)) : 
	  ((Storage_Manager<T> *) x.build_storage_manager(db_conn, (ondisk_storage_spec*) spec, hm));
	//skiing spec
	skiing_spec ski_spec;
	ski_spec.initial_resort = initial_resort;
	ski_spec.taoVal = tao_val;
	ski_spec.m_factor = m_factor;
	ski_spec.ratio = ratio;
	Skiing ski_obj(ski_spec);
	
	sgd = new Read_Specialized_Hazy_Sgd<T>(st_manager, inc_sgd, ski_obj, spec->strategy, low_water, high_water);
      }
    }
  }
  
  ///Server is initialized
  serv = new Hazy_Server();
  cout << "server started" << endl;
  
  while(1) {
    Hazy_Worker* hw = serv->acceptConnection();
    if (hw == NULL) {
      printf("Error at accept() : %s\n", strerror(errno));
      continue;
    }
    
    pthread_t t;
    worker_thread_info<T> *wit = new worker_thread_info<T>(hw, sgd, server_id);
    int iret = pthread_create(&t, NULL, worker_thread<T>, (void *) wit);
    if(iret != 0 ) {
      cerr << "pthread create failed " << iret << endl;
      break;
    }
    workers.push_back(t);
  }
  
  ///Server connection is closed
  serv->closeConnection();
  ///Database connection is closed
  db_conn->closeConnection();
}
