#include "hazy_message.h"
#include "vectors.h"
#include "postgres_util.h" 
#include "frontend_util.h"
#include "server_conn.h"
#include <iostream>
#include <string>
#include <math.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include "vectors.h"
#include "timer.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>

using namespace std;

/**
 * This class tests the server functionality
 * It simulates both server and client to check that we correctly send and receive the data
 */

const string close_conn_str = "close_conn";

void *server_thread_method (void *ptr) {
  cout << "in server thread" << endl;
  Hazy_Server *serv = new Hazy_Server();
  cout << "server started" << endl;
  
  while(1) {
    cout << "waiting to accept connection" << endl;
    Hazy_Worker *hw = serv->acceptConnection();
    if (hw == NULL) {
      printf("Error at accept() : %s\n", strerror(errno));
      continue;
    } 
    
    while (1) {
      void *data = hw->readData();
      if(strcmp((char *) data, close_conn_str.c_str()) == 0)
	break;
      if(hw->writeData(data, strlen((char *) data), false) < 0)
      	cout << "[Worker] error when writing to the client: " << strerror(errno);	
    }
    
    hw->closeClientConnection();
    break;
  }
  
  cout << "connection closed" << endl;
  serv->closeConnection();
  
  return NULL;
}

void test_single_connection(int sockfd, int dim, bool &success) {
  FVector fvector;
  // empty read for the ack?
  
  Timer t(true);
  for(int j = 0; j < dim; j ++)
    fvector.set(j, ((VFloat) drand48() * 2 - 1));
  
  int label = (int) (drand48() * 2);
  ostringstream ostr;
  ostr << label << " " << fvector;
  cout << " Vector constructed in " << t.stop() << endl;
  
  string msg = ostr.str().substr(0, ostr.str().size() - 1); //remove new line
  //string msg = "test";
  Timer send_timer(true);
  int  numOfBytesSent = write_wire_message(sockfd, msg.c_str(), strlen(msg.c_str()));
  if(numOfBytesSent < 0) {
    printf("[ERROR:Send]  %s\n", strerror(errno));
    return;
  }
  cout << "\t sent " << send_timer.stop() << endl;
  
  Timer recv_timer(true);
  buffer_handle buf = initialize_buffer();
  void *pData = read_wire_message(&buf, sockfd);
  cout << "\t recv " << recv_timer.stop() << endl;
  if(strcmp((char *) pData, msg.c_str()) == 0) {
    //cout << "match" << endl;
    /*cout << "match: read from server: " << buf.pData << endl;
      cout << "match: msg in client   : " << msg << endl;*/
  }
  else {
    cout << "non-match: read from server: " << buf.pData << endl;
    cout << "non-match: msg in client   : " << msg << endl;
    success = false;
    destroy_handle(buf);
    return;
  }
  destroy_handle(buf);
}

void test_communication_with_dim(int sockfd, int dim, bool &success) {
  success = true;
  int t = 0;
  while (t < 20) {
    test_single_connection(sockfd, dim, success);
    if(!success)
      return;
    
    t ++;
    //cout << "[test_communication_with_dim] dim: " << dim << ", end of iteration " << t << endl;
  }
}

void test_communication_with_dim_with_random(int sockfd, int upper_bound, bool &success) {
  int t = 0;
  int dim = rand() % upper_bound;
  while (t < 20) {
    int new_dim = dim - 10 * t;	//to test that when we read smaller data, there is no data from previous query
    test_single_connection(sockfd, new_dim, success);
    if(!success)
      return;
    
    t ++;
    //cout << "[test_communication_with_dim_with_random] dim: " << dim << ", end of iteration " << t << endl;
  }
}

void *client_thread_method (void *ptr) {
  cout << "in client thread" << endl;
  struct addrinfo *res = NULL;
  int sockfd = -1;
  struct addrinfo hints;
  
  if(sockfd < 0) {
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags  = AI_NUMERICSERV;
    if( getaddrinfo("localhost", "10003", &hints, &res) != 0 ) {
      res = NULL;
      cerr << "get addrinfo failed" << endl;
      return NULL;
    }
    
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    
    if(sockfd < 0) {
      sockfd = -1;
      cerr << "sock cannot be initialized" << endl;
      return NULL;
    }
    
    if(connect(sockfd, res->ai_addr, res->ai_addrlen) < 0) {
      sockfd = -1;
      cerr << "connect failed" << endl;
      return NULL;
    }
    int flag = 1;
    if(setsockopt( sockfd, IPPROTO_TCP, TCP_NODELAY, (char *)&flag, sizeof(flag) ) == -1) {
      printf("Couldn't setsockopt(TCP_NODELAY)\n");
      exit( EXIT_FAILURE );
    }
    
  }
  
  bool success = true;
  test_communication_with_dim(sockfd, 20, success);
  cout << "test_communication_with_dim(20): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim(sockfd, 100, success);
  cout << "test_communication_with_dim(100): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim(sockfd, 500, success);
  cout << "test_communication_with_dim(500): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim(sockfd, 1000, success);
  cout << "test_communication_with_dim(1000): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim(sockfd, 10000, success);
  cout << "test_communication_with_dim(10000): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim(sockfd, 100000, success);
  cout << "test_communication_with_dim(100000): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim_with_random(sockfd, 20, success);
  cout << "test_communication_with_dim_with_random(20): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim_with_random(sockfd, 100, success);
  cout << "test_communication_with_dim_with_random(100): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim_with_random(sockfd, 500, success);
  cout << "test_communication_with_dim_with_random(500): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim_with_random(sockfd, 1000, success);
  cout << "test_communication_with_dim_with_random(1000): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim_with_random(sockfd, 10000, success);
  cout << "test_communication_with_dim_with_random(10000): " << (success ? "successful" : "failed") << endl;
  test_communication_with_dim_with_random(sockfd, 100000, success);
  cout << "test_communication_with_dim_with_random(100000): " << (success ? "successful" : "failed") << endl;
  
  int  numOfBytesSent = write_wire_message(sockfd, close_conn_str.c_str(), strlen(close_conn_str.c_str()));
  if(numOfBytesSent < 0) {
    printf("[ERROR:Send]  %s\n", strerror(errno));
    return NULL;
  }
  
  return NULL;
}

int main(int argc, char **argv) {
  cout << "test initialized" << endl;
  pthread_t server_thread, client_thread;
  pthread_create(&server_thread, NULL, server_thread_method, (void *) NULL);
  
  cout << "server thread created" << endl;
  sleep(1);
  
  pthread_create(&client_thread, NULL, client_thread_method, (void *) NULL);	
  cout << "client thread created" << endl;
  
  pthread_join(client_thread, NULL);
  pthread_join(server_thread, NULL);
  
  
  return 0;
}
