#include "storage_manager.h"
#include "main_memory_storage_manager.h"
#include "globals.h"
#include <errno.h>
#include <string>
#include "frontend_util.h"
#include "hybrid_storage_manager.h"

using namespace std;

/**
 * Implements simple tests for main memory storage manager
 */

template<class T>
void test1(Storage_Manager<T> *sm, const char* filename,  struct hazy_model& hm) {
  sm->loadFromFile(hm, filename);
  cout << "Loaded file " << filename << endl;
  int c = 0;
  
  for(int i = 0; i < 2; i++) {
    sm->getEntityClass(2, c, hm);
    cout << " Entity 2 has class " << c << endl;
    double waste = 0;
    
    int n0 = 0;
    sm->getNumInClass(0, n0, hm, waste);
    cout << " Number in class: " << n0 << " waste: " << waste << endl;
    sm->incrementalUpdate(hm,waste);
    
    sm->resort(hm);
    cout << " Resort " << endl;
  } 
}

template <class T>
void test_spec(struct main_memory_storage_spec *s, struct model &m, char* filename) {
  struct hazy_model hm(m, s->strategy);
  hm.low_water = -1.0;
  hm.high_water = 1.0;
  Storage_Manager<T>* sm          = (Storage_Manager<T>*) new STLMainMemory_Storage_Manager<T>(s);
  cout << "Constructed the storage manager " << endl;
  test1(sm, filename, hm);
  cout << "Test Concluded" << endl;
}

int
main(int argc, char**argv) {
  opterr = 1;
  // Parameters
  int iDense  = 0;
  int dim     = 10;
  char *input_model_file = NULL, *filename = NULL;
  static struct extended_option long_options[] = {
    {"dense"    , no_argument, &iDense, 1, "if input contains dense vectors"},
    {"dim"      , required_argument, NULL, 'd', "the dimension of the model"},
    {"model"    , required_argument, NULL, 'm', "existing model input filename"},
    {NULL,0,NULL,0,0} 
  };
  char usage[] = "<entity_filename>";
  // end command line parameters
  int c = 0, option_index;
  while( (c = getopt_long(argc, argv, "m:", convert_extended_options(long_options), &option_index)) != -1) {
    switch(c) {      
    case 'd':
      dim = atoi(optarg);
      break;
    case 'm':
      input_model_file = optarg;
      break;
    case ':':
    case '?':
      print_usage(long_options, argv[0], usage);
    exit(-1);
    break;
    default:
      break;
    }
  }
  
  bool bDense = iDense > 0; // getopt insists on having ints for flags
  
  if(optind == argc - 1) {
    filename = argv[optind];
  } else {
    print_usage(long_options, argv[0], usage);
    exit(-1);
  }
  
  // Construct a fake model
  struct model m(dim,true);
  if(input_model_file) { 
    DEBUG_ONLY(cout << "Using a model from " << input_model_file << endl;);
    std::ifstream f(input_model_file);
    // should take in a model.
    f >> m;
  }
  std::cout << " dense=" << bDense << std::endl;
  
  main_memory_storage_spec *spec = new main_memory_storage_spec();
  spec->unique_id               = 0;
  spec->bRebuild                = true; 
  spec->strategy                = hazy_model::LAZY_NAIVE;
  
  std::cout << "LAZY NAIVE " << std::endl;
  if(bDense) { test_spec<FVector>(spec, m, filename); } else { test_spec<SVector>(spec, m, filename); }
  
  std::cout << "EAGER NAIVE " << std::endl;
  spec->strategy = hazy_model::EAGER_NAIVE;
  if(bDense) { test_spec<FVector>(spec, m, filename); } else { test_spec<SVector>(spec, m, filename); }
  

  std::cout << "LAZY HAZY " << std::endl;
  spec->strategy = hazy_model::LAZY_HAZY;
  if(bDense) { test_spec<FVector>(spec, m, filename); } else { test_spec<SVector>(spec, m, filename); }
  
  
  std::cout << "EAGER HAZY " << std::endl;
  spec->strategy = hazy_model::EAGER_HAZY;
  if(bDense) { test_spec<FVector>(spec, m, filename); } else { test_spec<SVector>(spec, m, filename); }
  
}
