#ifndef __globals_h
#define __globals_h

#ifdef __cplusplus
#include <assert.h>
#include <cstdlib>
#include <iostream>
#include <string>
#include <sstream>
#endif

/**
 * This class contains methods for various logging messages such as logging, debugging, verbose.
 */

///Debug mode
#ifdef __DEBUG
#define DEBUG_ONLY(x) x
#else
#define DEBUG_ONLY(x)
#endif


///Verbose mode
#ifdef __VERBOSE
#define VERBOSE_ONLY(x) x;
#else 
#define VERBOSE_ONLY(x);
#endif

///Logging mode
#ifdef __LOGGING
#define LOGGING_ONLY(x) x;
#else 
#define LOGGING_ONLY(x);
#endif

enum storage_manager_type {MAIN_MEMORY_SM, ON_DISK_SM, HYBRID_SM};

///TABLES
#define CLASSIFICATION_GENERAL_TABLE "hazy_cls_view_catalog"
#define FEATURE_IDS_TABLE_BASE_NAME "featureids"
#define EXTRACTED_FEATURES "extractedfeatures"
#define HAZY_ENTITY_TABLE_BASE_NAME "hazy_entity_table"
#define TRAINING_TABLE "training"

///EXTRACTED_FEATURES TABLE COL NAMES
#define EXTRACTED_FEATURES_FEATURE_VAL_COL "feature_value"

///Function Names
#define GET_STATS_FEATURES_FUNCTION_NAME "getstatsfeatures"
#define NORMALIZE_GET_M_FACTOR_FUNCTION_NAME "normalizegetm_factor"
#define CREATE_ENTITY_FEATURES_FUNCTION_NAME "create_entity_features"
#define PREPARE_TRAIN_FUNCTION_NAME "prepare_train"
#define DOT_PRDCT_CACHED_WASTE "dotprdct_cached_waste"
#define DOT_PRDCT_CACHED_WIHOUT_WASTE "dotprdct_cached_without_waste"

#define DOT_PRDCT_SPARSE_CACHED_WASTE "dotprdct_sparse_cached_waste"
#define DOT_PRDCT_SPARSE_CACHED_WITHOUT_WASTE "dotprdct_sparse_cached_without_waste"
#define DOT_PRDCT_DENSE_CACHED_WASTE "dotprdct_dense_cached_waste"
#define DOT_PRDCT_DENSE_CACHED_WITHOUT_WASTE "dotprdct_dense_cached_without_waste"

#define INITIALIZE_WASTE_COUNT "initialize_waste_count"
#define GET_WASTE_RATIO "get_waste_ratio"
#define CACHE_MODEL "cache_model_waste"

///Hazy Types
#define PAIR_TYPE "pair"
#define PAIR_ARRAY_TYPE "pair[]"
#define FLOAT8_TYPE "float8"
#define FLOAT8_ARRAY_TYPE "float8[]"

///CLASSIFICATION_GENERAL_TABLE FIELDS
#define VIEW_ID "view_id"
#define VIEW_NAME "view_name"
#define ENTITY_TABLE_NAME "entity_table_name"
#define LABELS_TABLE_NAME "labels_table"
#define EXAMPLE_TABLE_NAME "examples_table"
#define FEATURE_FUNCTION "feature_function"

//POSTGRES_CONSTANTS
//TABLES
#define PG_CLASS_TABLE "pg_class"

//PG_CLASS_TABLE COL
#define PG_CLASS_RELNAME_COL "relname"

#define MAX_PATH_LENGTH 1024

//DIR NAMES
#define INSTALLATION_DIR "installation"
#define TRIGGERS_DIR "triggers"
#define FUNCTIONS_DIR "functions"
#define UPDATE_TRIG_FILE_NAME "update_trig"
#define INPLACE_UPDATE_FNC_FILE_NAME "inplace"
#define BACKEND_FNC_FILE_NAME "all_backend_sparse_on_disk_functions"
#define BACKEND_FNC_NAME "backend_functions"
#define SOURCE_CODE_EXT "c"
#define OBJECT_FILE_EXT "o"
#define SHARED_OBJECT_FILE_EXT "so"
#define INCLUDE_DIR "include"
#define COMM_SERVER_DIR "communication_server"
#define SVM_CODE_DIR "svm_code"
#define S_LIBS_DIR "s_libs"

//executable create view name
#define CREATE_VIEW_EXEC "exec/hazy_create_view"

//conf file
#define HAZY_CONF_FILE "hazy.conf"
#define POSTGRES_PATH_LINE_IN_CONF_FILE "POSTGRES_PATH"
#define DATABASE_LINE_IN_CONF_FILE "Database"
#define STORAGE_MANAGER_LINE_IN_CONF_FILE "Storage_Manager"
#define USE_EPS_MAP_LINE_IN_CONF_FILE "Use_Eps_Map"
#define HAZY_STRATEGY_LINE_IN_CONF_FILE "Hazy_Strategy"
#define RESERVOIR_SIZE_LINE_IN_CONF_FILE "Reservoir_Size"
#define BUFFER_SIZE_LINE_IN_CONF_FILE "Buffer_Size"

//storage manager strings
#define MAIN_MEMORY_SM_IN_CONF_FILE "Main_Memory_SM"
#define ON_DISK_SM_IN_CONF_FILE "Ondisk_SM"
#define HYBRID_SM_IN_CONF_FILE "Hybrid_SM"

//hazy strategy strings
#define LAZY_NAIVE_IN_CONF_FILE "LAZY_NAIVE"
#define EAGER_NAIVE_IN_CONF_FILE "EAGER_NAIVE"
#define LAZY_HAZY_IN_CONF_FILE "LAZY_HAZY"
#define EAGER_HAZY_IN_CONF_FILE "EAGER_HAZY"


//postgres dir names
#define POSTGRES_INCLUDE "include"
#define POSTGRES_SERVER "server"
#define POSTGRES_LIB "lib"

//trigger functions
#define UPDATE_TRIG_FNC "update_trig"
#define UPDATE_TRIG_BIN_FNC "update_trig_bin"
#define UPDATE_TRIG_FNC_ON_TABLE "training_trig"

//compilers
#define GCC "gcc"
#define GPP "g++"

#endif
