/**
 * \file hashing.h
 */
#ifndef HASHING_H
#define HASHING_H

#ifdef __GNUC__
#include <ext/hash_map>
#else
#include <hash_map>
#endif

#include<iostream>

using namespace std;

namespace std
{
  using namespace __gnu_cxx;
}


template<typename T1, typename T2>
  
  /**
   * Hash map class
   */
  class Hazy_Hashmap {
 private:
  
 public:
  ///Default constructor
  Hazy_Hashmap() {}
  ///Hash map
  hash_map<T1,T2> map;
  
  /**
   * Inserts given pair of key and element to the map
   * 
   * \param[in] key Key of map in type T1
   * \param[in] element Value of key in type T2
   */
  void insertPair(T1 key, T2 element) {map[key] = element;}
  
  /**
   * Finds value of a key. If key cannot be found, then it returns 0(null)
   *
   * \param[in] key Key of map in type T1
   * \param[out] found boolean that is set to key is found or not
   *
   * \return value of key if found, else null 
   */
  T2 findKeyValue(T1 key, bool &found) 
  {
    typename hash_map<T1, T2>::iterator it = map.find(key);
    
    if(it != map.end()) {
      found = true;
      return it->second;
    }
    else {
      found = false;
      return 0;
    }
  };
  
  /**
   * Updates value of given key
   * 
   * \param[in] key Key of map in type T1
   * \param[in] element Value of key in type T2
   * 
   */
  void replaceValue(T1 key, T2 element) {map[key] = element;}
  
  /**
   * Clears map
   */
  void clear() {map.clear();};
  
  /**
   * Gives size of hash map
   * 
   * \return size of hash map
   */
  int size() {return map.size();}
};

#endif
