#ifndef __server_conn_h
#define __server_conn_h

/**
 * \file server_conn.h
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "low_level_net.h"

#define CON_NUM 10
#define MAX_IPV4_SIZE 15
//#define BUFF_SIZE 100000

#define WRITE_SIZE 4096
class Hazy_Worker;

/**
 * This class creates socket server connection and accepts connection from clients
 */

class Hazy_Server {
 private:
  ///server socket id
  int sockfd;
  ///buffer for communication
  buffer_handle bh;
  struct addrinfo hints, *servinfo;
  struct sockaddr_in their_addr;
  socklen_t sin_size;
  char ip[MAX_IPV4_SIZE]; // max would be lets say 111.111.111.111 = 3×4 ( nums ) + 3×1 ( dots ) = 15
  int rc;
 public:
  /**
   * Default constructor
   */
  Hazy_Server();
  /**
   * Initializes connection
   */
  void inetConn();
  /**
   * Accepts connection
   */
  Hazy_Worker* acceptConnection();
  /**
   * Closes server connection
   */
  void closeConnection();
};

class Hazy_Worker {
  ///Buffer for read/writes
  buffer_handle bh;
  ///Client connection socket id
  int fd; // this the connection
  ///Buffer size
  int buf_size;
  
 public:
  /**
   * Initializes buffer and sets client socket id
   * \param[in] fd client socket id
   */
  Hazy_Worker(int fd) {
    bh = initialize_buffer();
    this->fd = fd;
  }
  /**
   * Hazy_Worker destructor that destructs buffer
   */
  ~Hazy_Worker() { destroy_handle(bh); } 
  
  /**
   * Reads data from client
   */
  void* readData();
  /**
   * Writes data to the client
   * \param[in] data data pointer to be written to the client
   * \param[in] size size of the data pointer
   * \param[in] intData boolean value shows whether data is integer data or not
   */
  int writeData(const void *data, int size, bool intData);
  /**
   * Closes client connection
   */
  void closeClientConnection();
};

#endif
