/**
 * \file sgd_model.h
 */
#ifndef __sgd_model_h
#define __sgd_model_h

#include "vectors.h"
#include <cmath>
#include "globals.h"
#include <algorithm>
#include "model.h"
#include "incremental_sgd.h"

///Class for maintaining svm model
class SvmSgd {
 private:
  int _dim;
  model _current;
  ///number of predictions
  int ex;
  ///number of correct prediction during testing
  int match;
  ///Updates old model
  Incremental_SGD *sgd;
 public:
  /**
   * SvmSgd constructor
   * \param[in] dim dimension of the model
   * \param[in] lambda lambda of the sgd learning
   */
  SvmSgd(int dim, double lambda);
  /**
   * Scales model with wscale
   */
  void scale();
  /**
   * Retrieves model learnt
   * \return model learnt
   */	
  struct model& get_model() { return _current;} 
  /**
   * Tests given example in svm format and check whether prediction is correct or not
   */
  void test_example(char *newfeature);
  /**
   * Returns accuracy on test data
   * \return accuracy on test data
   */
  double getAccuracy();
  /**
   * Checks whether given example is classified correctly or not
   * \return whether example is correctly classified by the current model or not
   */
  bool isExampleCorrectlyClassified(std::string newfeature);
};

#endif

