db_name=$1
current_location=$2

psql $db_name -c "

DROP TYPE IF EXISTS pair CASCADE;

CREATE TYPE pair;
CREATE FUNCTION pair_in(cstring)
    RETURNS pair
    AS '$current_location/common_functions.so'
    LANGUAGE C IMMUTABLE STRICT;

CREATE FUNCTION pair_out(pair)
    RETURNS cstring
    AS '$current_location/common_functions.so'
    LANGUAGE C IMMUTABLE STRICT;

CREATE FUNCTION pair_recv(internal)
   RETURNS pair
    AS '$current_location/common_functions.so'
   LANGUAGE C IMMUTABLE STRICT;

CREATE FUNCTION pair_send(pair)
   RETURNS bytea
    AS '$current_location/common_functions.so'
   LANGUAGE C IMMUTABLE STRICT;

CREATE TYPE pair (
   internallength = 16,
   input = pair_in,
   output = pair_out,
   receive = pair_recv,
   send = pair_send,
   alignment = double
);
";

psql $db_name -c "

DROP AGGREGATE IF EXISTS array_accum(anyelement); DROP AGGREGATE IF EXISTS group_concat(text);

create aggregate array_accum (
	sfunc = array_append,
	basetype = anyelement,
	stype = anyarray,
	initcond = '{}'
);

CREATE OR REPLACE FUNCTION _group_concat(text, text)
RETURNS text AS \$\$
	SELECT CASE
		WHEN \$2 IS NULL THEN \$1
		WHEN \$1 IS NULL THEN \$2
		ELSE \$1 operator(pg_catalog.||) ',' operator(pg_catalog.||) \$2
END
\$\$ IMMUTABLE LANGUAGE SQL;

CREATE AGGREGATE group_concat (
	BASETYPE = text,
	SFUNC = _group_concat,
	STYPE = text
);
";

psql $db_name -c "DROP TYPE IF EXISTS __auxT CASCADE; CREATE TYPE __auxT AS (wordf text, tff float);";
psql $db_name -c "DROP AGGREGATE IF EXISTS sumText(text); CREATE AGGREGATE sumText( BASETYPE = text, SFUNC = textcat, STYPE = text, INITCOND = '' );";
psql $db_name -c "DROP FUNCTION IF EXISTS auxF(TEXT, INTEGER, INTEGER); CREATE OR REPLACE FUNCTION auxF(TEXT, INTEGER, INTEGER) RETURNS SETOF __auxT AS '$current_location/backend_functions.so', 'auxF' LANGUAGE C IMMUTABLE STRICT;";
psql $db_name -c "DROP FUNCTION IF EXISTS dotprdct_cached_waste(pair[]); CREATE OR REPLACE FUNCTION dotprdct_cached_waste(pair[]) RETURNS FLOAT8 AS '$current_location/backend_functions.so', 'dotprdct_sparse_cached_waste' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS dotprdct_cached_without_waste(pair[]); CREATE OR REPLACE FUNCTION dotprdct_cached_without_waste(pair[]) RETURNS FLOAT8 AS '$current_location/backend_functions.so', 'dotprdct_sparse_cached_without_waste' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS dotprdct_cached_waste(float8[]); CREATE OR REPLACE FUNCTION dotprdct_cached_waste(float8[]) RETURNS FLOAT8 AS '$current_location/backend_functions.so', 'dotprdct_dense_cached_waste' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS dotprdct_cached_without_waste(float8[]); CREATE OR REPLACE FUNCTION dotprdct_cached_without_waste(float8[]) RETURNS FLOAT8 AS '$current_location/backend_functions.so', 'dotprdct_dense_cached_without_waste' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS initialize_waste_count(); CREATE OR REPLACE FUNCTION initialize_waste_count() RETURNS VOID AS '$current_location/backend_functions.so', 'initialize_waste_count' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS get_waste_ratio(); CREATE OR REPLACE FUNCTION get_waste_ratio() RETURNS FLOAT8 AS '$current_location/backend_functions.so', 'get_waste_ratio' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS cache_model_waste(FLOAT8[], FLOAT8); CREATE OR REPLACE FUNCTION cache_model_waste(FLOAT8[], FLOAT8) RETURNS VOID AS '$current_location/backend_functions.so', 'cache_model_waste' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS getEntityLabel(INTEGER) CASCADE; CREATE OR REPLACE FUNCTION getEntityLabel(INTEGER) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getEntityLabel' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS function_noop(INTEGER) CASCADE; CREATE OR REPLACE FUNCTION function_noop(INTEGER) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'function_noop' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS function_noop_withserver(INTEGER) CASCADE; CREATE OR REPLACE FUNCTION function_noop_withserver(INTEGER) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'function_noop_withserver' LANGUAGE C;";
psql $db_name -c "DROP FUNCTION IF EXISTS getView(); CREATE OR REPLACE FUNCTION getView() RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getView' LANGUAGE C IMMUTABLE STRICT;";
psql $db_name -c "DROP FUNCTION IF EXISTS getServerId(); CREATE OR REPLACE FUNCTION getServerId() RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getServerId' LANGUAGE C IMMUTABLE STRICT;";
psql $db_name -c "DROP FUNCTION IF EXISTS closeServerConn(); CREATE OR REPLACE FUNCTION closeServerConn() RETURNS INTEGER AS '$current_location/frontend_functions.so', 'closeServerConn' LANGUAGE C IMMUTABLE STRICT;";
psql $db_name -c "DROP FUNCTION IF EXISTS getEntityLabelWithSQL(FLOAT8, TEXT) CASCADE; CREATE OR REPLACE FUNCTION getEntityLabelWithSQL(FLOAT8, TEXT) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getEntityLabelWithSQL_Text' LANGUAGE C";
psql $db_name -c "DROP FUNCTION IF EXISTS getEntityLabelWithSQL(FLOAT8, FLOAT8[]) CASCADE; CREATE OR REPLACE FUNCTION getEntityLabelWithSQL(FLOAT8, FLOAT8[]) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getEntityLabelWithSQL_Dense' LANGUAGE C";
psql $db_name -c "DROP FUNCTION IF EXISTS getEntityLabelWithSQL(FLOAT8, pair[]) CASCADE; CREATE OR REPLACE FUNCTION getEntityLabelWithSQL(FLOAT8, pair[]) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getEntityLabelWithSQL_Sparse' LANGUAGE C";
psql $db_name -c "DROP FUNCTION IF EXISTS getEntityLabelWithSQLNOEPS(TEXT) CASCADE; CREATE OR REPLACE FUNCTION getEntityLabelWithSQLNOEPS(TEXT) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getEntityLabelWithSQLNOEPS_Text' LANGUAGE C";
psql $db_name -c "DROP FUNCTION IF EXISTS getEntityLabelWithSQLNOEPS(FLOAT8[]) CASCADE; CREATE OR REPLACE FUNCTION getEntityLabelWithSQLNOEPS(FLOAT8[]) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getEntityLabelWithSQLNOEPS_Dense' LANGUAGE C";
psql $db_name -c "DROP FUNCTION IF EXISTS getEntityLabelWithSQLNOEPS(pair[]) CASCADE; CREATE OR REPLACE FUNCTION getEntityLabelWithSQLNOEPS(pair[]) RETURNS INTEGER AS '$current_location/frontend_functions.so', 'getEntityLabelWithSQLNOEPS_Sparse' LANGUAGE C";
psql $db_name -c "DROP FUNCTION IF EXISTS update_external_table_inplace(sm_test_external_table, bool) CASCADE; CREATE OR REPLACE FUNCTION update_external_table_inplace(sm_test_external_table, bool) RETURNS BOOLEAN AS '$current_location/inplace.so', 'update_hack' LANGUAGE C STRICT;";

psql $db_name < functions/delete_table_view.sql
