#include "postgres.h"
#include "funcapi.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include <string.h>
#include "utils/array.h"
#include "catalog/pg_type.h"
#include <pthread.h>
#include "executor/spi.h"       /* this is what you need to work with SPI */
#include "commands/trigger.h"   /* ... and triggers */

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include <sys/socket.h>
#include <netdb.h>

#include <assert.h>

#include "fmgr.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

/**
 * Methods for newly created type (Pair) are implemented here
 */

PG_FUNCTION_INFO_V1(pair_in);

typedef struct Pair {
  int         x;
  double      y;
} Pair;

Datum
pair_in(PG_FUNCTION_ARGS)
{
  char       *str = PG_GETARG_CSTRING(0);
  int  	  x;
  double      y;
  Pair       *result;
  
  if (sscanf(str, " ( %ld , %lf )", &x, &y) != 2)
    ereport(ERROR,
	    (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
	     errmsg("invalid input syntax for complex: \"%s\"",
		    str)));
  
  result = (Pair *) palloc(sizeof(Pair));
  result->x = x;
  result->y = y;
  PG_RETURN_POINTER(result);
}

PG_FUNCTION_INFO_V1(pair_out);

Datum
pair_out(PG_FUNCTION_ARGS)
{
  Pair       *pair = (Pair *) PG_GETARG_POINTER(0);
  char       *result;
  
  result = (char *) palloc(100);
  snprintf(result, 100, "(%d,%g)", pair->x, pair->y);
  PG_RETURN_CSTRING(result);
}

PG_FUNCTION_INFO_V1(pair_recv);

Datum
pair_recv(PG_FUNCTION_ARGS)
{
  StringInfo  buf = (StringInfo) PG_GETARG_POINTER(0);
  Pair       *result;
  
  result = (Pair *) palloc(sizeof(Pair));
  result->x = pq_getmsgint(buf, 4);
  result->y = pq_getmsgfloat8(buf);
  PG_RETURN_POINTER(result);
}

PG_FUNCTION_INFO_V1(pair_send);

Datum
pair_send(PG_FUNCTION_ARGS)
{
  Pair          *pair = (Pair *) PG_GETARG_POINTER(0);
  StringInfoData buf;
  
  pq_begintypsend(&buf);
  pq_sendint(&buf, pair->x, 4);
  pq_sendfloat8(&buf, pair->y);
  PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}
