#include "postgres.h"

#include "funcapi.h"

#include "access/heapam.h"
#include "access/hio.h"
#include "access/multixact.h"
#include "access/relscan.h"
#include "access/sysattr.h"
#include "access/transam.h"
#include "access/tuptoaster.h"
#include "access/valid.h"
#include "access/visibilitymap.h"
#include "access/xact.h"
#include "access/xlogutils.h"
#include "catalog/catalog.h"
#include "catalog/namespace.h"
#include "miscadmin.h"
#include "pgstat.h"
#include "storage/bufmgr.h"
#include "storage/freespace.h"
#include "storage/lmgr.h"
#include "storage/procarray.h"
#include "storage/smgr.h"
// #include "storage/standby.h"
#include "utils/datum.h"
#include "utils/inval.h"
#include "utils/lsyscache.h"
#include "utils/relcache.h"
#include "utils/snapmgr.h"
#include "utils/syscache.h"
#include "utils/tqual.h"
#include "utils/builtins.h"

#include "executor/spi.h"
#include "commands/trigger.h"

extern Datum update_hack0(HeapTupleHeader htup, bool value);
PG_MODULE_MAGIC;


static char szDebug0[2048];

// Properties of the table
static char *relation_name0 = "demo_labeled_papers";
// Assume the schema
// (id,eps,class)
#define TABLE_ARITY 3
static bool  bNull0[TABLE_ARITY]      = {false, false, false};
static bool  bIsReplace0[TABLE_ARITY] = {false, false, true};
static Datum vals0[TABLE_ARITY];
static int   UPDATE_INDEX0 = 2; 

Datum
update_hack0(HeapTupleHeader htup, bool value) {
  
  // Currently we are hacked to take the relation name as a string.
  int i = 0;
  // Fetch the object id for the tuple
  Oid          tid   = HeapTupleHeaderGetOid(htup);
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, "Entering Hack Two, Got Values\n");)
#endif
    
    
    RangeVar   *relvar;
  Relation	rel;
  bool isnull;
  List	   *relname_list;
  
  relname_list = stringToQualifiedNameList(relation_name0); // HACK
  relvar = makeRangeVarFromNameList(relname_list);
  rel = relation_openrv(relvar, AccessShareLock);
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, "Found Relation Names");)   
#endif
    
    // Now we form the tuple (we don't want anything null)
    HeapTupleData tup;
  tup.t_data = htup; //(HeapTupleHeader) PageGetItem(page, lp);
  // Fix up datum
  tup.t_len  = HeapTupleHeaderGetDatumLength(htup);// ItemIdGetLength(lp);
  sprintf(szDebug0, "Setup length=%d\n", tup.t_len);
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, szDebug0);)  
#endif
    
    // Get the relation oid.
    tup.t_tableOid = RelationGetRelid(rel);
  
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, "Got Relid...");)  ;
#endif
  tup.t_self = htup->t_ctid;
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, "Old Tuple Constructed...");)  ;
#endif
  
  // Modify the tuple
  // HeapTupleHeaderGetNatts
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, "Getting Value"););
#endif
  
  // This bit of code sets up the values for below. 
  // This looks a bit unnecessary: Why copy out?
  for(i = 0; i < TABLE_ARITY; i++) {
    vals0[i] = GetAttributeByNum(htup, i+1, &isnull);
  }
  // STORE THE UPDATE VALUE
  vals0[UPDATE_INDEX0] = value;
  
#ifdef DEBUGGING
  DEBUG_ONLY(sprintf(szDebug0, "Got Datums: %d -- %d\n", DatumGetInt32(vals0[0]), value););
  DEBUG_ONLY(elog(INFO, szDebug0););
#endif
  
  // This tuple description should likely be cached.
  TupleDesc tupdesc   = RelationNameGetTupleDesc(relation_name0);
#ifdef DEBUGGING
  DEBUG_ONLY(sprintf(szDebug0, "Got Tuple Description: %d", tupdesc->natts);)
    DEBUG_ONLY(elog(INFO, szDebug0);)
#endif
    
    // Assert that the tuple description and how many attributes we
    // think are presenta are the same, i.e.,
    // assert(tupdesc->natts == TABLE_ARITY);
#ifdef DEBUGGING
    DEBUG_ONLY(elog(INFO, "About to modify the tuple...");)
#endif
    HeapTuple ntup = heap_modify_tuple(&tup, tupdesc, vals0, bNull0, bIsReplace0);
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, "New tuple built"););
#endif

  // Do the store
  heap_inplace_update(rel, ntup);
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, "STORED!"););
#endif
  
  // Cleanup
  //ReleaseTupleDesc(tupdesc);
  relation_close( rel, AccessShareLock);
#ifdef DEBUGGING
  DEBUG_ONLY(elog(INFO, "Done"));
#endif
  PG_RETURN_BOOL(false);
}

