#include "hazy_message.h"
#include <pthread.h>
#include "postgres.h"
#include "funcapi.h"
#include "executor/executor.h"  /* for GetAttributeByName() */
#include <string.h>
#include "utils/array.h"
#include "catalog/pg_type.h"
#include "executor/spi.h"       /* this is what you need to work with SPI */
#include "commands/trigger.h"   /* ... and triggers */

#include "fmgr.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

const int MAX_STRING_SIZE=16840;

extern Datum trigf(PG_FUNCTION_ARGS);

PG_FUNCTION_INFO_V1(trig_noop);

/**
 * When new training example is inserted, does nothing for experimental purposes
 */
Datum
trig_noop(PG_FUNCTION_ARGS)
{
  HeapTuple   rettuple;
  TriggerData *trigdata = (TriggerData *) fcinfo->context;
  TupleDesc   tupdesc;
  bool        checknull = false;
  bool        isnull;
  int         ret, i;
  // make sure it's called as a trigger at all
  if (!CALLED_AS_TRIGGER(fcinfo))
    elog(ERROR, "trigf: not called by trigger manager");
  
  if (TRIGGER_FIRED_BY_INSERT(trigdata->tg_event)) {
    rettuple = trigdata->tg_trigtuple;
    checknull = true;
    
    tupdesc = trigdata->tg_relation->rd_att;
    
    char* relname = SPI_getrelname(trigdata->tg_relation);
    int id =  atoi(SPI_getvalue(rettuple, tupdesc, 1));
    
    // connect to SPI manager
    if ((ret = SPI_connect()) < 0)
      elog(INFO, "trigf (fired): SPI_connect returned %d", ret);
    
    pfree(relname);
    
    if (checknull) {   
      SPI_getbinval(rettuple, tupdesc, 1, &isnull);
      if (isnull)
	rettuple = NULL;
    }
    SPI_finish();
  }
  return PointerGetDatum(rettuple);
}

