#include <sys/time.h>
#include <string.h>
#include <fstream>
#include <stdio.h>
#include "libpq-fe.h"
#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include <map>
#include <vector>
#include <cassert>
#include <cstdlib>
#include <cmath>
#include <algorithm>
#include <iterator>
#include "pthread.h"
#include "timer.h"

using namespace std;


struct prepared_stmt {
  string _stmt_name;
  char** _params;
  int _nParams;
  prepared_stmt(int nParams) {
    _params = new char*[nParams];
    _nParams = nParams;
  }
  ~prepared_stmt() {
    for(int i = 0; i < _nParams;i++) {
      free(_params[i]);
    }
    delete _params;
  }
};

void
exit_nicely(PGconn *conn)
{
  PQfinish(conn);
  return;
}

PGconn*
prepare_connection(string db_name, string prepared_file_name, int server_id) {
  string conninfo = "dbname = " + db_name;
  PGresult *res;
  PGconn* conn = PQconnectdb(conninfo.c_str());

  if (PQstatus(conn) == CONNECTION_BAD)
    {
      fprintf(stderr, "Connection to database '%s' failed.\n", db_name.c_str());
      fprintf(stderr, "%s", PQerrorMessage(conn));
    } 
  
  int server_id_from_query = -1;
  
  
  while(server_id_from_query != server_id) {
    res = PQexec(conn, "SELECT getServerId();");
    server_id_from_query = atoi(PQgetvalue(res, 0, 0));
    cout << "\tgot id="<< server_id_from_query << " wanted " << server_id << endl;
    if(server_id_from_query != server_id) { sleep(5); }
    PQclear(res);
  }
  
  cerr << "it seems server started" << endl;




  ifstream preparefile(prepared_file_name.c_str());
  if(preparefile.is_open()) {
    string line;
    while(!preparefile.eof() && getline(preparefile, line)) {
      res = PQexec(conn, line.c_str());
      if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK))
	{
	  fprintf(stderr, "%s", PQerrorMessage(conn));
	  fprintf(stderr, "prepared statement %s could not be created. It might be because table for prepared statement does not exist\n", line.c_str());
	}
      PQclear(res);
    }
  }
  
	
  res = PQexec(conn, "START TRANSACTION ISOLATION LEVEL READ COMMITTED;");
  PQclear(res);
  return conn;
}


// The worker thread
struct thread_info {
  PGconn* _conn;
  vector<struct prepared_stmt*> &_stmts;
  thread_info(PGconn *conn,  vector<struct prepared_stmt*> &stmts) : _conn(conn), _stmts(stmts) { } 
};

void*
worker_thread(void *parms) {
  struct thread_info *ti = (struct thread_info*) parms;
  PGconn* _conn = ti->_conn;
  PGresult *res;

  vector<prepared_stmt*> &examples = ti->_stmts;
  typedef vector<prepared_stmt*>::iterator prep_stmt_iterator;
  for(prep_stmt_iterator i = examples.begin(); 
      i != examples.end(); i ++) {
    res = PQexecPrepared(_conn, (*i)->_stmt_name.c_str(), (*i)->_nParams, (*i)->_params, NULL, NULL, 1);
        
    if (!res || (PQresultStatus(res) != PGRES_COMMAND_OK && PQresultStatus(res) != PGRES_TUPLES_OK)) {
      fprintf(stderr, "PROBLEM in insert query in measure, error: %s\n", PQerrorMessage(_conn));
      PQclear(res);
      exit_nicely(_conn);
    }
    PQclear(res);
  }
  
  res = PQexec(_conn, "END;");
  PQclear(res);
  return NULL;
}

int 
main(int argc, char** argv) {
  PGresult *res = NULL;
  string query; 
  string line = "";
  string db_name  = argv[1];
  ifstream myfile (argv[2]);
  ofstream wrfile (argv[3]);	
  string prepared_file_name = argv[4];
  int numOfParameters = atoi(argv[5]);
  int server_id = atoi(argv[6]);
  int nThreads = atoi(argv[7]);

  vector<prepared_stmt*> _prepped;
  
  //read sql file
  if (myfile.is_open()) {
    while (! myfile.eof() && getline (myfile,line)) {
      stringstream str;
      str << line;
      string word;
      prepared_stmt *ps = new prepared_stmt(numOfParameters);
      getline(str, word, ';');
      ps->_stmt_name = strdup(word.c_str());
      for(int j = 0; j < numOfParameters; j ++) {
	getline(str, word, ';');
	ps->_params[j] = strdup(word.c_str());
      }
      _prepped.push_back(ps);
    }
  }
  cout << "Prepared all statements" << endl;

	// Setup the connection objects
  PGconn **conns = new PGconn*[nThreads];
  for(int i = 0; i < nThreads; i++) {
    conns[i] = prepare_connection(db_name, prepared_file_name, server_id);
  }
  
  cout << "All connections setup." << endl;
  vector<struct prepared_stmt*>*p_stmts = new vector<struct prepared_stmt*>[nThreads];
  // We assign round robin
  int nExamples = _prepped.size();
  for(int i = 0; i < nExamples; i++) {
    int index = i % nThreads;
    p_stmts[index].push_back(_prepped[i]);
  }
  
  cout << "Examples divided up" << endl;
  
  pthread_t threads[nThreads];
  // **** START EXPERIMENT ****
  Timer t(true);
  for(int i = 0; i < nThreads; i++) {	  
    int ret = pthread_create(&threads[i], NULL, worker_thread,  (void*) (new thread_info(conns[i], p_stmts[i]))); 
    if(ret != 0  ) { 
      cout << "Error in pthread_create: " << ret << endl;
      exit(-1); 
    }	
  }
  for(int i = 0; i < nThreads; i++) {	  
    int ret = pthread_join(threads[i], NULL);
    if(ret != 0) {
      cout << "Error in pthread_join: " << ret << endl;
      exit(-1); 
	  }
  }

  wrfile << t.stop() << endl;
  // **** END EXPERIMENT ****

  if(nThreads > 0) {
    res = PQexec(conns[0], "SELECT closeServerConn()");
    if(atoi(PQgetvalue(res, 0, 0)) == 1)
      cout << "server successfully closed" << endl;
  }

  delete [] p_stmts;

  PQclear(res);	
  sleep(100);
  for(int i = 0; i < nThreads; i++) {
    PQfinish(conns[i]);
  }
  
	wrfile.close();
  return 0;	
}
