#include "storage_manager.h"
#include "globals.h"
#include "timer.h"
#include <fstream>
//Storage_Manager::Storage_Manager() { }

/**
 * Destructor
 */
template<class T>
Storage_Manager<T>::~Storage_Manager() { }

// Expects the following format
// 1 SVECTOR
/**
 * Loads entities from the given file
 * \param[in] m hazy model
 * \param[in] szFileName file name
 */
template<class T>
void 
Storage_Manager<T>::
loadFromFile(struct hazy_model& hm, const char* szFileName) {  
  DEBUG_ONLY(Timer t1(true););
  std::ifstream our_file(szFileName);
  int nExamples = 0;
  if(our_file.is_open()) {
    int id = 0;
    std::string line;
    while(! our_file.eof() && getline(our_file, line) ) {
      std::istringstream f(line + "\n");
      T x;
      f >> id >> x;
      VERBOSE_ONLY(std::cout << "got: " << id << " - " << x << " " << line << std::endl;); 
      insertEntity(hm, id, x);
      nExamples++;
    }
  }
  DEBUG_ONLY(std::cout << "Loaded " << nExamples << " tuples in " << t1.stop() << "s" << std::endl;);
}

/**
 * Loads entities from the database
 * \param[in] m hazy model
 */
template<class T>
int
Storage_Manager<T>::loadFromDatabase(struct hazy_model &hm) {
  if(db_conn == NULL) {
    fprintf(stderr, "Error: Entities cannot be loaded from database. Hazy_Database is NULL");
    return -1;
  }
  std::vector< sParse_IDvec<T> > _out;
  std::string query = "SELECT id, feature_vector FROM " + entity_table_name + ";";
  int ret_val = db_conn->postgresTupleParser(query.c_str(), sParse_IDvec<T>::parse_tuple, _out);
  if(ret_val < 0) {
    fprintf(stderr, "error in load from database method in storage manager: %s\n", entity_table_name.c_str());
    return -1;
  }
  
  for(unsigned int i = 0; i < _out.size(); i ++) {
    insertEntity(hm, _out.at(i).id, _out.at(i).f);
  }
  
  return ret_val;
}
