#ifndef _model_h
#define _model_h
#include <sstream>
#include <cstdlib>
#include <iostream>

#include "vectors.h"
#include "globals.h"

/**
 * This structure represents svm model that contains feature vector, wscale, bias, dimension and learning rate
 */
struct model {
  FVector w;
  double wscale, bias;
  int dim;
  double t;
  
  /**
   * Constructor that initializes a model with the given dimension
   * \param[in] dim dimension of the model
   */
model(int dim) : w(dim) { wscale = 1; bias = 0; this->dim = dim;}
  /**
   * Default constructor that initializes a model
   */
model() : w(0) {wscale = 1; bias = 0; this->dim = dim;} 
  /**
   * Constructor that initializes a model with the given dimension and optinal random noise
   * \param[in] dim dimension of the model
   * \param[in] random_noise whether model should be filled randomly or not
   */
model(int dim, bool random_noise) : w(dim) {
  wscale = 1;
  bias = drand48();
  if(random_noise) {
    srand(10);
    for(int i = 0; i < dim; i++) {
      double v = drand48() - 0.5;
      w.set(i, v);
    }
  }
  this->dim = dim;
}

  void resetModel() {
    wscale = 1;
    bias = 0;
    w.clear();
    w.resize(dim);
  }
};
/**
 * Parses given input stream to create a model
 */
std::istream& operator>>(std::istream &f, model &m);
/**
 * Writes model to the output stream
 */
std::ostream& operator<<(std::ostream &f, const model &m);
#endif
