/**
 * \file sgd_model.cxx
 */
#include "sgd_model.h"

/**
 * Constructor that initializes SvmSgd
 * \param[in] dim dimension of the model
 * \param[in] l lambda
 */
SvmSgd::SvmSgd(int dim, double l)
  : _current(dim), _last_sort_model(dim)
{
  _dim = dim;
  ex = 0;
  match = 0;
}

/**
 * Scales model with wscale
 */
void
SvmSgd::scale() {
  _current.w.scale(_current.wscale);
  _current.wscale = 1;
}

/**
 * Returns accuracy on test data
 * \return accuracy on test data
 */
double
SvmSgd::getAccuracy() {
  if(ex == 0)
    return 0;
  return (double) match * 100 / ex;
}

/**
 * Tests given example in svm format and check whether prediction is correct or not
 * \param[in] featurechar feature vector of the entity in svm format
 */
void
SvmSgd::test_example(char *featurechar) {
  std::string featurestr = featurechar;
  bool isCorrectlyClassified = isExampleCorrectlyClassified(featurestr);
  
  if(isCorrectlyClassified)
    match ++;
  ex ++;
}

/**
 * Checks whether given example is classified correctly or not
 * \param[in] featurestr feature vector of the entity in svm format
 * \return whether entity is correctly classified by the current model or not
 */
bool
SvmSgd::isExampleCorrectlyClassified(std::string featurestr) {
  SVector x;
  double y = x.loadString(featurestr);
  double wx = dot(_current.w,x) * _current.wscale;
  double product = wx + _current.bias;
  double z = y * product;
  
  return z > 0;
}
