#include <cstdlib>
#include <iostream>
#include <string>
#include "vectors.h"
#include "postgres_util.h"
#include <sstream>

using namespace std;

/**
 * This class tests dot product and waste ratio functions in postgres_backend functions
 */
int main(int argc, char **argv) {
	srand(5);
	int numSVectors = 10;
	int dim = 20;
	int max_svector_dim = 10;

	string dbname = argv[1];

	Hazy_Database *db = new Hazy_Database(dbname);

	FVector w(dim);
	for(int i = 0; i < w.size(); i ++)
		w.set(i, ((VFloat) drand48() * 2 - 1));	//to make it negative, too

	cout.precision(16);

	double bias = drand48() * 2 -1;

	FVector fVectorList[numSVectors];

	SVector sVectorList[numSVectors];

	for(int i = 0; i < numSVectors; i ++) {
		int svector_dim = int(drand48() * max_svector_dim);
		while(svector_dim == 0)
			svector_dim = int(drand48() * max_svector_dim);

		for(int j = 0; j < svector_dim; j ++) {
			sVectorList[i].set(int(drand48() * dim), ((VFloat) drand48() * 2 -1));
		}
		for(int j = 0; j < w.size(); j ++)
			fVectorList[i].set(j, ((VFloat) drand48() * 2 - 1));
	}


	string fvec_db_str = "";
	fvector_to_dbstring(w, fvec_db_str);

	//dense waste test
	ostringstream query;
	std::ios::fmtflags oldflags = query.flags();
  std::streamsize oldprec = query.precision();
  query << std::scientific << std::setprecision(sizeof(double)==4 ? 7 : 16);

	query << "SELECT initialize_waste_count();";
  query << "SELECT cache_model_waste(" << fvec_db_str << ", " << bias << ");";

  query.precision(oldprec);
  query.flags(oldflags);

  db->execute_statement_msg(query.str().c_str(), "msg");

  int non_waste = 0;

  cout.precision(16);

	for(int i = 0; i < numSVectors; i ++) {
		double prediction = dot(w, fVectorList[i]) + bias;
		if(prediction <= 0)
			non_waste ++;

		ostringstream new_query;
		string fvector_str = "";
		fvector_to_dbstring(fVectorList[i], fvector_str);
		new_query << "SELECT dotprdct_cached_waste(" << fvector_str << ");";
		double prediction_result = 0;
		db->execute_query_msg_double(new_query.str().c_str(), "dotprdct", prediction_result);
    if(prediction != prediction_result) {
      cout << "prediction: " << prediction << endl;
      cout << "prediction result: " << prediction_result << endl;
    }
	}

	double waste_ratio = (double) non_waste / (double) numSVectors;
  ostringstream waste_query;
  waste_query << "SELECT get_waste_ratio();";
  double waste_ratio_in_db = 0;
  db->execute_query_msg_double(waste_query.str().c_str(), "waste ratio", waste_ratio_in_db);

  if(waste_ratio != waste_ratio_in_db) {
    cout << "waste ratio: " << waste_ratio << endl;
    cout << "waste ratio in db: " << waste_ratio_in_db << endl;
  }

	cout << "end of dense waste test" << endl;

	//dense without waste test
	query.str("");
  oldflags = query.flags();
  oldprec = query.precision();
  query << std::scientific << std::setprecision(sizeof(double)==4 ? 7 : 16);

  query << "SELECT initialize_waste_count();";
  query << "SELECT cache_model_waste(" << fvec_db_str << ", " << bias << ");";

  query.precision(oldprec);
  query.flags(oldflags);

  db->execute_statement_msg(query.str().c_str(), "msg");

  non_waste = 0;

  cout.precision(16);

  for(int i = 0; i < numSVectors; i ++) {
    double prediction = dot(w, fVectorList[i]) + bias;
    if(prediction <= 0)
      non_waste ++;

    ostringstream new_query;
    string fvector_str = "";
    fvector_to_dbstring(fVectorList[i], fvector_str);
    new_query << "SELECT dotprdct_cached_without_waste(" << fvector_str << ");";
    double prediction_result = 0;
    db->execute_query_msg_double(new_query.str().c_str(), "dotprdct", prediction_result);
    if(prediction != prediction_result) {
      cout << "prediction: " << prediction << endl;
      cout << "prediction result: " << prediction_result << endl;
    }
  }

  waste_ratio = 0;
	waste_query.str("");
  waste_query << "SELECT get_waste_ratio();";
  waste_ratio_in_db = 0;
  db->execute_query_msg_double(waste_query.str().c_str(), "waste ratio", waste_ratio_in_db);

  if(waste_ratio != waste_ratio_in_db) {
    cout << "waste ratio: " << waste_ratio << endl;
    cout << "waste ratio in db: " << waste_ratio_in_db << endl;
  }

	cout << "end of dense without waste test" << endl;

	//sparse waste test
	query.str("");
	oldflags = query.flags();
  oldprec = query.precision();
  query << std::scientific << std::setprecision(sizeof(double)==4 ? 7 : 16);

	query << "SELECT initialize_waste_count();";
	query << "SELECT cache_model_waste(" << fvec_db_str << ", " << bias << ");";

	query.precision(oldprec);
	query.flags(oldflags);

	db->execute_statement_msg(query.str().c_str(), "msg");
	
	non_waste = 0;

	cout.precision(16);

	for(int i = 0; i < numSVectors; i ++) {
		double prediction = dot(w, sVectorList[i]) + bias;
		if(prediction <= 0)
			non_waste ++;

		ostringstream new_query;
		string svector_str = "";
		svector_to_dbstring(sVectorList[i], svector_str);
		new_query << "SELECT dotprdct_cached_waste(" << svector_str << ");";
		double prediction_result = 0;
		db->execute_query_msg_double(new_query.str().c_str(), "dotprdct", prediction_result);
		if(prediction != prediction_result) {
			cout << "prediction: " << prediction << endl;
			cout << "prediction result: " << prediction_result << endl;
		}
	}

	waste_ratio = (double) non_waste / (double) numSVectors;
	waste_query.str("");
	waste_query << "SELECT get_waste_ratio();";
	waste_ratio_in_db = 0;
	db->execute_query_msg_double(waste_query.str().c_str(), "waste ratio", waste_ratio_in_db);

	if(waste_ratio != waste_ratio_in_db) {
		cout << "waste ratio: " << waste_ratio << endl;
		cout << "waste ratio in db: " << waste_ratio_in_db << endl;
	}

	cout << "end of sparse waste test" << endl;

	//sparse without waste test
	query.str("");
	query << "SELECT initialize_waste_count();";
  query << "SELECT cache_model_waste(" << fvec_db_str << ", " << bias << ");";

  query.precision(oldprec);
  query.flags(oldflags);

  db->execute_statement_msg(query.str().c_str(), "msg");

  non_waste = 0;

  cout.precision(16);

  for(int i = 0; i < numSVectors; i ++) {
    double prediction = dot(w, sVectorList[i]) + bias;
    if(prediction <= 0)
      non_waste ++;

    ostringstream new_query;
    string svector_str = "";
    svector_to_dbstring(sVectorList[i], svector_str);
    new_query << "SELECT dotprdct_cached_without_waste(" << svector_str << ");";
    double prediction_result = 0;
    db->execute_query_msg_double(new_query.str().c_str(), "dotprdct", prediction_result);
    if(prediction != prediction_result) {
      cout << "prediction: " << prediction << endl;
      cout << "prediction result: " << prediction_result << endl;
    }
  }

  waste_ratio = 0;
  waste_query.str("");
  waste_query << "SELECT get_waste_ratio();";
  waste_ratio_in_db = 0;
  db->execute_query_msg_double(waste_query.str().c_str(), "waste ratio", waste_ratio_in_db);

  if(waste_ratio != waste_ratio_in_db) {
    cout << "waste ratio: " << waste_ratio << endl;
    cout << "waste ratio in db: " << waste_ratio_in_db << endl;
  }

}
