#ifndef _lvector_h_
#define _lvector_h_ 1
#include "vectors.h"
//#include "p_pair_vectors.h"
// PSQL includes
#include "libpq-fe.h"
#include "postgres_util.h"
#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "utils/array.h"
#include "utils/lsyscache.h"
#include "catalog/pg_type.h"

#include <ios>
#include <vector>

template <class T>
struct _full_entity_entry {
  int k;
  T x;
  double eps;
  bool label;

  typedef _full_entity_entry<T> this_vector;

  _full_entity_entry<T> operator=(const _full_entity_entry<FVector>&them) {
    x = them.x;
    label = them.label;
    return *this;
  }
  _full_entity_entry(T _x, int _l){ x = _x; label = _l; }
  _full_entity_entry(){  }

  static int Parse_Vec(PGresult *res, int index, this_vector &c) {
    c.k = atoi(PQgetvalue(res, index, 0));
    std::string f_vec = PQgetvalue(res, index, 1);
    c.eps = atof(PQgetvalue(res, index, 2));
    c.label = c.eps > 0;
    int r = c.x.fromPSQL(f_vec);

    return r;
  }


  static int binary_Parse_Vec(PGresult *res, int index, this_vector &c) {
    static int buffer_size = 0;
    
    assert(PQfformat(res,0) == 1 && PQfformat(res,1) == 1);
    assert(PQftype(res,0) == INT4OID);

    c.k = (int) unpack_int32(PQgetvalue(res, index, 0));
    // Now unpack the array.
    char *array_data = PQgetvalue(res, index, 1);
    int blen = PQgetlength(res, index, 1);
    assert(blen > 20);
    // ******************************************
    // see arrayfuncs.c:array_send around 1480
    int ndim     = unpack_int32(array_data);
    int has_null = unpack_int32(array_data + 4);
    assert(ndim == 1 && has_null == 0);
    int oid      = unpack_int32(array_data + 8);

    if (oid == FLOAT8OID) {
      int nDoubles = unpack_int32(array_data + 12);
    
      assert(ndim == 1 && has_null == 0);
      VERBOSE_ONLY(std::cout << " Found a vector of length " << blen << std::endl;);
      VERBOSE_ONLY(std::cout << "\t ndims=" << ndim << " has_null=" << has_null << " oid=" << oid << " isFloat8?" << (oid==FLOAT8OID) << std::endl;);
      /* VERBOSE_ONLY(std::cout << "\t dims=" << nDoubles << " lbound=" << lbound << std::endl;); */

      char *binary_data   = array_data + 20;
      buffer_size = nDoubles;
      // format is length:value
      for(int i = 0; i < nDoubles ; i++) {
	int len             = unpack_int32(binary_data);
	c.x.set(i, unpack_double(binary_data+sizeof(int)));
	binary_data       += len + sizeof(int);
      }
    }
    else {
      int nPairs = unpack_int32(array_data + 12);
      //int lbound   = unpack_int32(array_data + 16);
      assert(ndim == 1 && has_null == 0);
      VERBOSE_ONLY(std::cout << " Found a vector of length " << blen << std::endl;);
      VERBOSE_ONLY(std::cout << "\t ndims=" << ndim << " has_null=" << has_null << " oid=" << oid << " isFloat8? " << (oid==FLOAT8OID) << std::endl;);
      VERBOSE_ONLY(std::cout << "\t dims=" << nPairs << std::endl;);

      char *binary_data   = array_data + 20;
      buffer_size = nPairs;

      // format is length:value
      for(int i = 0; i < nPairs ; i++) {
	int len             = unpack_int32(binary_data);
	c.x.set(unpack_int32(binary_data+sizeof(int)), unpack_double(binary_data+2*sizeof(int)));
	binary_data       += len + sizeof(int);
      }
    }

    c.eps = (double) unpack_double(PQgetvalue(res, index, 2));
    c.label = c.eps > 0;

    return 0;
}

  static bool _full_entity_compare(const _full_entity_entry &x, const _full_entity_entry &y) {
    return x.eps < y.eps;
  }
};

#endif
