#include "parse_conf_file.h"

char* retrieve_hazy_conf_file_path() {
  char buf[MAX_PATH_LENGTH];
  bzero(buf, MAX_PATH_LENGTH);
  readlink("/proc/self/exe", buf, sizeof(buf) - 1);
  //buf is ../exec/executable_name

  char *pch;
  pch = strrchr(buf, '/');
  int slash_index = pch - buf;
  int conf_file_str_length = (strlen(buf) + strlen(HAZY_CONF_FILE) + 10);
  char *path_to_conf_file = (char *) malloc(sizeof(char) * conf_file_str_length);
  bzero(path_to_conf_file, conf_file_str_length);
  strncpy(path_to_conf_file, buf, slash_index);
  strcat(path_to_conf_file, "/../../");
  strcat(path_to_conf_file, HAZY_CONF_FILE);

  return path_to_conf_file;
}

char* retrieve_hazy_dirs_path() {
  char buf[MAX_PATH_LENGTH];
  bzero(buf, MAX_PATH_LENGTH);
  readlink("/proc/self/exe", buf, sizeof(buf) - 1);
  
  //buf is ../exec/executable_name
  
  char *pch;
  pch = strrchr(buf, '/');
  int slash_index = pch - buf;
  int hazy_dirs_str_length = (strlen(buf) + 10);
  char *path_to_hazy_dirs = (char *) malloc(sizeof(char) * hazy_dirs_str_length);
  bzero(path_to_hazy_dirs, hazy_dirs_str_length);
  strncpy(path_to_hazy_dirs, buf, slash_index);
  strcat(path_to_hazy_dirs, "/../");

  return path_to_hazy_dirs;
}

#ifdef __cplusplus
int parse_postgres_path_from_conf_file(std::string conf_file_path, std::string &postgres_path) {
  std::string line = "";
  bool postgres_line_found = false;

  std::ifstream hazy_conf_file(conf_file_path.c_str());
  if(hazy_conf_file.is_open()) {
    while(hazy_conf_file.good()) {
      getline(hazy_conf_file, line);
      if(line.find(POSTGRES_PATH_LINE_IN_CONF_FILE) != std::string::npos) {
	size_t eq_found = line.find_first_of("=");
	if(eq_found == std::string::npos)
	  return POSTGRES_PATH_PARSE_PROBLEM;
	postgres_path = line.substr(eq_found + 1);
	postgres_line_found = true;
      }
    }
  }
  else {
    std::cerr << conf_file_path << " file could not be found" << std::endl;
    return FILE_NOT_FOUND;
  }
  
  if(!postgres_line_found)
    return POSTGRES_LINE_NOT_FOUND;
  
  return SUCCESSFUL_PARSE;
}

int parse_dbname_from_conf_file(std::string conf_file_path, std::string &db_name) {
  std::string line = "";
  bool dbname_line_found = false;

  std::ifstream hazy_conf_file(conf_file_path.c_str());
  if(hazy_conf_file.is_open()) {
    while(hazy_conf_file.good()) {
      getline(hazy_conf_file, line);
      if(line.find(DATABASE_LINE_IN_CONF_FILE) != std::string::npos) {
	size_t eq_found = line.find_first_of("=");
	if(eq_found == std::string::npos)
	  return DBNAME_PARSE_PROBLEM;

	db_name = line.substr(eq_found + 1);
	dbname_line_found = true;
      }
    }
  }
  else {
    std::cerr << conf_file_path << " file could not be found" << std::endl;
    return FILE_NOT_FOUND;
  }

  if(!dbname_line_found)
    return DBNAME_LINE_NOT_FOUND;

  return SUCCESSFUL_PARSE;
}

int parse_hazy_settings(std::string conf_file_path, enum storage_manager_type &sm, int &useepsmap, 
			 hazy_model::strategy &hazy_strategy, int &reservoir_size, int &buffer_size) {
  std::string line = "";
  bool sm_line_processed = false;
  bool useepsmap_line_processed = false;
  bool hazy_strategy_line_processed = false;
  bool reservoir_size_line_processed = false;
  bool buffer_size_line_processed = false;

  std::ifstream hazy_conf_file(conf_file_path.c_str());
  if(hazy_conf_file.is_open()) {
    while(hazy_conf_file.good()) {
      getline(hazy_conf_file, line);
      size_t eq_found = line.find_first_of("=");
      std::string parsed_str = line.substr(eq_found + 1);

      if(line.find(STORAGE_MANAGER_LINE_IN_CONF_FILE) != std::string::npos) {
	if(eq_found == std::string::npos)
	  return SM_PARSE_PROBLEM;
	//check strings
	if(parsed_str.compare(MAIN_MEMORY_SM_IN_CONF_FILE) == 0)
	  sm = MAIN_MEMORY_SM;
	else if(parsed_str.compare(ON_DISK_SM_IN_CONF_FILE) == 0)
	  sm = ON_DISK_SM;
	else if(parsed_str.compare(HYBRID_SM_IN_CONF_FILE) == 0)
	  sm = HYBRID_SM;
	else
	  return SM_PARSE_PROBLEM;

	sm_line_processed = true;
      }
      else if(line.find(USE_EPS_MAP_LINE_IN_CONF_FILE) != std::string::npos) {
	if(eq_found == std::string::npos)
	  return USEEPS_PARSE_PROBLEM;
	useepsmap = atoi(parsed_str.c_str());
	useepsmap_line_processed = true;
      }
      else if(line.find(HAZY_STRATEGY_LINE_IN_CONF_FILE) != std::string::npos) {
	if(eq_found == std::string::npos)
	  return HAZY_ST_PARSE_PROBLEM;

	//check strings
	if(parsed_str.compare(LAZY_NAIVE_IN_CONF_FILE) == 0)
	  hazy_strategy = hazy_model::LAZY_NAIVE;
	else if(parsed_str.compare(EAGER_NAIVE_IN_CONF_FILE) == 0)
	  hazy_strategy = hazy_model::EAGER_NAIVE;
	else if(parsed_str.compare(LAZY_HAZY_IN_CONF_FILE) == 0)
	  hazy_strategy = hazy_model::LAZY_HAZY;
	else if(parsed_str.compare(EAGER_HAZY_IN_CONF_FILE) == 0)
	  hazy_strategy = hazy_model::EAGER_HAZY;
	else
	  return HAZY_ST_PARSE_PROBLEM;

	hazy_strategy_line_processed = true;

      }
      else if(line.find(RESERVOIR_SIZE_LINE_IN_CONF_FILE) != std::string::npos) {
	if(eq_found == std::string::npos)
	  return RESERVOIR_SIZE_PARSE_PROBLEM;
	reservoir_size = atoi(parsed_str.c_str());
	reservoir_size_line_processed = true;
      }
      else if(line.find(BUFFER_SIZE_LINE_IN_CONF_FILE) != std::string::npos) {
	if(eq_found == std::string::npos)
	  return BUFFER_SIZE_PARSE_PROBLEM;
	buffer_size = atoi(parsed_str.c_str());
	buffer_size_line_processed = true;
      }
    }
  }
  else {
    std::cerr << conf_file_path << " file could not be found" << std::endl;
    return FILE_NOT_FOUND;
  }

  if(!sm_line_processed)
    return SM_PARSE_LINE_NOT_FOUND;
  if(!useepsmap_line_processed)
    return USEEPS_LINE_NOT_FOUND;
  if(!hazy_strategy_line_processed)
    return HAZY_ST_LINE_NOT_FOUND;
  if(!reservoir_size_line_processed)
    return RESERVOIR_LINE_NOT_FOUND;
  if(!buffer_size_line_processed)
    return BUFFER_SIZE_LINE_NOT_FOUND;

  return SUCCESSFUL_PARSE;
}

void handle_success_code(int parse_success_code) {
  switch(parse_success_code) {
  case SUCCESSFUL_PARSE:
    break;
  case POSTGRES_PATH_PARSE_PROBLEM:
    std::cerr << POSTGRES_PATH_LINE_IN_CONF_FILE << " could not be parsed in the conf file" << std::endl;
    std::cerr << "e.g. " << POSTGRES_PATH_LINE_IN_CONF_FILE << "=/path/to/postgres" << std::endl;
    break;
  case DBNAME_PARSE_PROBLEM:
    std::cerr << DATABASE_LINE_IN_CONF_FILE << " could not be parsed in the conf file" << std::endl;
    std::cerr << "e.g. " << DATABASE_LINE_IN_CONF_FILE << "=test_db" << std::endl;
    break;
  case SM_PARSE_PROBLEM:
    std::cerr << STORAGE_MANAGER_LINE_IN_CONF_FILE << " could not be parsed in the conf file" << std::endl;
    std::cerr << "e.g. " << STORAGE_MANAGER_LINE_IN_CONF_FILE << "=Main_Memory_SM" << std::endl;
    break;
  case USEEPS_PARSE_PROBLEM:
    std::cerr << USE_EPS_MAP_LINE_IN_CONF_FILE << " could not be parsed in the conf file" << std::endl;
    std::cerr << "e.g. " << USE_EPS_MAP_LINE_IN_CONF_FILE << "=0" << std::endl;
    break;
  case HAZY_ST_PARSE_PROBLEM:
    std::cerr << HAZY_STRATEGY_LINE_IN_CONF_FILE << " could not be parsed in the conf file" << std::endl;
    std::cerr << "e.g. " << HAZY_STRATEGY_LINE_IN_CONF_FILE << "=EAGER_HAZY" << std::endl;
    break;
  case RESERVOIR_SIZE_PARSE_PROBLEM:
    std::cerr << RESERVOIR_SIZE_LINE_IN_CONF_FILE << " could not be parsed in the conf file" << std::endl;
    std::cerr << "e.g. " << RESERVOIR_SIZE_LINE_IN_CONF_FILE << "=1000" << std::endl;
    break;
  case BUFFER_SIZE_PARSE_PROBLEM:
    std::cerr << BUFFER_SIZE_LINE_IN_CONF_FILE << " could not be parsed in the conf file" << std::endl;
    std::cerr << "e.g. " << BUFFER_SIZE_LINE_IN_CONF_FILE << "=2000" << std::endl;
    break;
  case FILE_NOT_FOUND:
    std::cerr << HAZY_CONF_FILE << " could not be found" << std::endl;
    break;
  case POSTGRES_LINE_NOT_FOUND:
    std::cerr << POSTGRES_PATH_LINE_IN_CONF_FILE << " could not be found in the conf file" << std::endl;
    std::cerr << "e.g. " << POSTGRES_PATH_LINE_IN_CONF_FILE << "=/path/to/postgres" << std::endl;
    break;
  case DBNAME_LINE_NOT_FOUND:
    std::cerr << DATABASE_LINE_IN_CONF_FILE << " could not be found in the conf file" << std::endl;
    std::cerr << "e.g. " << DATABASE_LINE_IN_CONF_FILE << "=test_db" << std::endl;
    break;
  case SM_PARSE_LINE_NOT_FOUND:
    std::cerr << STORAGE_MANAGER_LINE_IN_CONF_FILE << " could not be found in the conf file" << std::endl;
    std::cerr << STORAGE_MANAGER_LINE_IN_CONF_FILE << " could not be parsed in the conf file" << std::endl;
    break;
  case USEEPS_LINE_NOT_FOUND:
    std::cerr << USE_EPS_MAP_LINE_IN_CONF_FILE << " could not be found in the conf file" << std::endl;
    std::cerr << "e.g. " << USE_EPS_MAP_LINE_IN_CONF_FILE << "=0" << std::endl;
    break;
  case HAZY_ST_LINE_NOT_FOUND:
    std::cerr << HAZY_STRATEGY_LINE_IN_CONF_FILE << " could not be found in the conf file" << std::endl;
    std::cerr << "e.g. " << HAZY_STRATEGY_LINE_IN_CONF_FILE << "=EAGER_HAZY" << std::endl;
    break;
  case RESERVOIR_LINE_NOT_FOUND:
    std::cerr << RESERVOIR_SIZE_LINE_IN_CONF_FILE << " could not be found in the conf file" << std::endl;
    std::cerr << "e.g. " << RESERVOIR_SIZE_LINE_IN_CONF_FILE << "=1000" << std::endl;
    break;
  case BUFFER_SIZE_LINE_NOT_FOUND:
    std::cerr << POSTGRES_PATH_LINE_IN_CONF_FILE << " could not be found in the conf file" << std::endl;
    std::cerr << "e.g. " << BUFFER_SIZE_LINE_IN_CONF_FILE << "=2000" << std::endl;
    break;
  default:
    std::cerr << "";
  }
}

#endif
