/**
 * \file timer.cxx
 */
#include "timer.h"

/**
 * Default constructor
 */
Timer::Timer() { bStart = false; } 
/**
 * Constructor that starts timer when called
 * \param[in] bStart when this is true, then time is started when constructor is called
 */
Timer::Timer(bool bStart) { this->bStart = bStart ; start(); } 
/**
 * Starts timer
 */
void Timer::start() { bStart = true; gettimeofday(&_start, NULL);}
/**
 * Stops timer and returns elapsed time
 * \return elapsed time
 */
double Timer::stop() { gettimeofday(&_stop, NULL); return getElapsedTime();}
/**
 * Returns elapsed time
 * \return elapsed time
 */
double 
Timer::getElapsedTime() {
  // Assert that we started 
  assert( bStart ); 
  timeval res;
  timersub(&_stop, &_start, &res);
  return res.tv_sec + res.tv_usec*1e-6;	
}
